import pytest
from unittest.mock import patch

from hera.data import Constant, DataLabel, DEFAULT_DATA_START, Label, Program, Settings
from hera.debugger import debug, Debugger, Shell
from hera.debugger.debugger import reverse_lookup_label
from hera.loader import load_program


@pytest.fixture
def shell():
    settings = Settings(color=False)
    return Shell(Debugger(SAMPLE_PROGRAM, settings), settings)


@pytest.fixture
def debugger():
    return Debugger(SAMPLE_PROGRAM, Settings())


def load_shell(program):
    settings = Settings(mode="debug")
    return Shell(Debugger(load_program(program, settings), settings), settings)


SAMPLE_PROGRAM = load_program(
    """\
// A comment
CONSTANT(N, 3)

SET(R1, N)
SET(R2, 39)
LABEL(add)
ADD(R3, R1, R2)
HALT()
""",
    Settings(mode="debug"),
)


def test_handle_break_prints_breakpoints(shell, capsys):
    shell.debugger.breakpoints[4] = "main.hera:7"
    shell.handle_command("break")

    assert capsys.readouterr().out == "main.hera:7\n"


def test_handle_break_prints_breakpoints_with_no_breakpoints_set(shell, capsys):
    shell.handle_command("break")

    assert capsys.readouterr().out == "No breakpoints set.\n"


def test_handle_break_sets_breakpoint(shell, capsys):
    assert len(shell.debugger.breakpoints) == 0

    shell.handle_command("break 5")

    assert len(shell.debugger.breakpoints) == 1
    assert 2 in shell.debugger.breakpoints
    assert shell.debugger.breakpoints[2] == "<string>:5"

    assert capsys.readouterr().out == "Breakpoint set in file <string>, line 5.\n"


def test_handle_break_with_dot(shell, capsys):
    shell.handle_command("break .")

    assert len(shell.debugger.breakpoints) == 1
    assert 0 in shell.debugger.breakpoints
    assert shell.debugger.breakpoints[0] == "<string>:4"

    assert capsys.readouterr().out == "Breakpoint set in file <string>, line 4.\n"


def test_handle_break_with_multiple_files(capsys):
    shell = load_shell('NOP()\n#include "test/assets/include/lib/add.hera"')
    shell.handle_command("break test/assets/include/lib/add.hera:1")

    assert len(shell.debugger.breakpoints) == 1
    assert 1 in shell.debugger.breakpoints
    assert shell.debugger.breakpoints[1] == "test/assets/include/lib/add.hera:1"

    assert (
        capsys.readouterr().out
        == "Breakpoint set in file test/assets/include/lib/add.hera, line 1.\n"
    )

    shell.handle_command("break 1")

    assert len(shell.debugger.breakpoints) == 2
    assert 0 in shell.debugger.breakpoints
    assert shell.debugger.breakpoints[0] == "<string>:1"

    assert capsys.readouterr().out == "Breakpoint set in file <string>, line 1.\n"


def test_handle_break_with_multiple_files_again(capsys):
    shell = load_shell('#include "test/assets/include/lib/add.hera"\nNOP()')

    shell.handle_command("break 1")

    assert len(shell.debugger.breakpoints) == 1
    assert 0 in shell.debugger.breakpoints

    captured = capsys.readouterr().out
    assert (
        captured == "Breakpoint set in file test/assets/include/lib/add.hera, line 1.\n"
    )


def test_handle_break_with_invalid_location(shell, capsys):
    shell.handle_command("break 1")

    assert len(shell.debugger.breakpoints) == 0
    assert capsys.readouterr().out == "Error: could not find corresponding line.\n"


def test_handle_break_with_unparseable_breakpoint(shell, capsys):
    shell.handle_command("break $$$")

    assert len(shell.debugger.breakpoints) == 0
    assert capsys.readouterr().out == "Error: could not locate label `$$$`.\n"


def test_handle_break_with_too_many_args(shell, capsys):
    shell.handle_command("break 1 2 3")

    assert len(shell.debugger.breakpoints) == 0
    assert capsys.readouterr().out == "break takes zero or one arguments.\n"


def test_handle_break_abbreviated(shell):
    with patch("hera.debugger.shell.Shell.handle_break") as mock_handle_break:
        shell.handle_command("b 7")
        assert mock_handle_break.call_count == 1

        args, kwargs = mock_handle_break.call_args
        assert len(args) == 1
        assert args[0] == ["7"]
        assert len(kwargs) == 0


def test_handle_clear(shell, capsys):
    shell.handle_command("break 4")
    capsys.readouterr()

    shell.handle_command("clear 4")

    assert len(shell.debugger.breakpoints) == 0

    captured = capsys.readouterr().out
    assert captured == "Cleared breakpoint in file <string>, line 4.\n"


def test_handle_clear_only_clears_one(shell):
    shell.handle_command("break 4")
    shell.handle_command("break 5")

    shell.handle_command("clear 4")

    assert len(shell.debugger.breakpoints) == 1
    assert 2 in shell.debugger.breakpoints


def test_handle_clear_with_multiple_args(shell, capsys):
    shell.handle_command("break 4")
    shell.handle_command("break 5")
    shell.handle_command("break 7")
    capsys.readouterr()

    shell.handle_command("clear 4 5")

    assert len(shell.debugger.breakpoints) == 1
    assert 4 in shell.debugger.breakpoints

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
Cleared breakpoint in file <string>, line 4.
Cleared breakpoint in file <string>, line 5.
"""
    )


def test_handle_clear_with_no_breakpoint(shell, capsys):
    shell.handle_command("clear 4")

    captured = capsys.readouterr().out
    assert captured == "No breakpoint at that location.\n"


def test_handle_clear_with_star_arg(shell, capsys):
    shell.handle_command("break 4")
    shell.handle_command("break 5")
    capsys.readouterr()

    shell.handle_command("clear *")

    assert len(shell.debugger.breakpoints) == 0

    captured = capsys.readouterr().out
    assert captured == "Cleared all breakpoints.\n"


def test_handle_clear_with_star_and_other_args(shell, capsys):
    shell.handle_command("break 4")
    shell.handle_command("break 5")
    capsys.readouterr()

    shell.handle_command("clear 4 * ???")

    assert len(shell.debugger.breakpoints) == 0

    captured = capsys.readouterr().out
    assert captured == "Cleared all breakpoints.\n"


def test_handle_clear_with_bad_arg(shell, capsys):
    shell.handle_command("clear ???")

    captured = capsys.readouterr().out
    assert captured == "Error: could not locate label `???`.\n"


def test_handle_clear_with_too_few_args(shell, capsys):
    shell.handle_command("clear")

    assert capsys.readouterr().out == "clear takes one or more arguments.\n"


def test_handle_clear_abbreviated(shell):
    with patch("hera.debugger.shell.Shell.handle_clear") as mock_handle_clear:
        shell.handle_command("cl 7")
        assert mock_handle_clear.call_count == 1

        args, kwargs = mock_handle_clear.call_args
        assert len(args) == 1
        assert args[0] == ["7"]
        assert len(kwargs) == 0


def test_handle_next(shell):
    assert shell.debugger.vm.registers[1] == 0
    assert shell.debugger.vm.pc == 0

    shell.handle_command("next")

    assert shell.debugger.vm.registers[1] == 3
    assert shell.debugger.vm.pc == 2


def test_handle_next_with_argument(shell):
    shell.handle_command("next 2")

    assert shell.debugger.vm.registers[1] == 3
    assert shell.debugger.vm.registers[2] == 39
    assert shell.debugger.vm.pc == 4


def test_handle_next_with_unparseable_argument(shell, capsys):
    shell.handle_command("next a")

    assert shell.debugger.vm.pc == 0
    assert capsys.readouterr().out == "Could not parse argument to next.\n"


def test_handle_next_with_HALT(shell, capsys):
    # Last instruction of SAMPLE_PROGRAM is a HALT operation.
    shell.debugger.vm.pc = 5

    shell.handle_command("next")

    assert shell.debugger.vm.pc == 5
    assert capsys.readouterr().out == "Program has finished executing.\n"


def test_handle_next_with_function_call(capsys):
    shell = load_shell(
        """\
SET(R1, 4)
CALL(FP_alt, plus_two)
SET(R2, 5)
HALT()

LABEL(plus_two)
  INC(R1, 2)
  RETURN(FP_alt, PC_ret)
"""
    )
    shell.handle_command("n")
    capsys.readouterr()
    shell.handle_command("next")

    assert shell.debugger.vm.pc == 5
    assert shell.debugger.vm.registers[1] == 6
    assert shell.debugger.vm.registers[2] == 0

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
[<string>]

    2  CALL(FP_alt, plus_two)
->  3  SET(R2, 5)
    4  HALT()
"""
    )


def test_handle_next_with_recursive_function(capsys):
    shell = load_shell(
        """\
SET(R1, 9)
CALL(FP_alt, rec)
HALT()

LABEL(rec)
  MOVE(FP, SP)
  INC(SP, 2)
  STORE(PC_ret, 0, FP)
  STORE(FP_alt, 1, FP)

  CMP(R1, R0)
  BZ(end)
  INC(R2, 1)
  DEC(R1, 1)
  CALL(FP_alt, rec)
  LABEL(end)

  LOAD(PC_ret, 0, FP)
  LOAD(FP_alt, 1, FP)
  DEC(SP, 2)
  RETURN(FP_alt, PC_ret)
"""
    )
    shell.handle_command("n")
    capsys.readouterr()
    shell.handle_command("next")

    assert shell.debugger.vm.pc == 5
    assert shell.debugger.vm.registers[1] == 0
    assert shell.debugger.vm.registers[2] == 9

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
[<string>]

     2  CALL(FP_alt, rec)
->   3  HALT()
     4
"""
    )


def test_handle_next_over_function_with_breakpoint(capsys):
    shell = load_shell(
        """\
CALL(FP_alt, foo)
HALT()

LABEL(foo)
  RETURN(FP_alt, PC_ret)
  """
    )
    shell.handle_command("break foo")
    capsys.readouterr()
    shell.handle_command("next")

    assert shell.debugger.vm.pc == 4

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
[<string>]

    4  LABEL(foo)
->  5    RETURN(FP_alt, PC_ret)
    6
"""
    )


def test_handle_next_after_end_of_program(shell, capsys):
    shell.debugger.vm.pc = 9000

    shell.handle_command("next")

    assert shell.debugger.vm.pc == 9000
    assert capsys.readouterr().out == "Program has finished executing.\n"


def test_handle_next_with_too_many_args(shell, capsys):
    shell.handle_command("next 10 11")

    assert capsys.readouterr().out == "next takes zero or one arguments.\n"


def test_handle_next_abbreviated(shell):
    with patch("hera.debugger.shell.Shell.handle_next") as mock_handle_next:
        shell.handle_command("n")
        assert mock_handle_next.call_count == 1


def test_handle_continue_with_breakpoint(shell):
    shell.debugger.breakpoints[4] = ""

    shell.handle_command("continue")

    vm = shell.debugger.vm
    assert vm.registers[1] == 3
    assert vm.registers[2] == 39
    assert vm.registers[3] == 0
    assert vm.pc == 4

    # Make sure continuing again doesn't loop on the same instruction.
    shell.handle_command("continue")
    assert vm.pc == 5


def test_handle_continue_without_breakpoint(shell, capsys):
    shell.handle_command("continue")

    vm = shell.debugger.vm
    assert vm.registers[1] == 3
    assert vm.registers[2] == 39
    assert vm.registers[3] == 42
    assert vm.pc == 5
    assert capsys.readouterr().out == "Program has finished executing.\n"


def test_handle_continue_with_too_many_args(shell, capsys):
    shell.handle_command("continue 10")

    assert capsys.readouterr().out == "continue takes no arguments.\n"


def test_handle_continue_abbreviated(shell):
    with patch("hera.debugger.shell.Shell.handle_continue") as mock_handle_continue:
        shell.handle_command("c")
        assert mock_handle_continue.call_count == 1


def test_handle_execute(shell):
    shell.handle_command("execute SET(R7, 42)")

    assert shell.debugger.vm.pc == 0
    assert shell.debugger.vm.registers[7] == 42


def test_handle_execute_with_branch(shell, capsys):
    shell.handle_command("execute BRR(10)")

    assert shell.debugger.vm.pc == 0
    assert capsys.readouterr().out == "execute cannot take branching operations.\n"


def test_handle_execute_with_data_statement(shell, capsys):
    shell.handle_command("execute INTEGER(42)")

    assert shell.debugger.vm.pc == 0
    assert capsys.readouterr().out == "execute cannot take data statements.\n"


def test_handle_execute_with_label(shell, capsys):
    shell.handle_command("execute LABEL(l)")

    assert shell.debugger.vm.pc == 0
    assert capsys.readouterr().out == "execute cannot take labels.\n"


def test_handle_execute_with_invalid_op(shell, capsys):
    shell.handle_command("execute ADD(R1, R2)")

    assert (
        capsys.readouterr().err
        == """\
Error: too few args to ADD (expected 3), line 1 col 1 of <string>

  ADD(R1, R2)
  ^

"""
    )


def test_handle_execute_with_no_args(shell, capsys):
    shell.handle_command("execute")

    assert capsys.readouterr().out == "execute takes one argument.\n"


def test_handle_goto_with_line_number(shell):
    shell.handle_command("goto 7")

    assert shell.debugger.vm.registers[1] == 0
    assert shell.debugger.vm.registers[2] == 0
    assert shell.debugger.vm.pc == 4


def test_handle_goto_with_label(shell):
    shell.handle_command("goto add")

    assert shell.debugger.vm.registers[1] == 0
    assert shell.debugger.vm.registers[2] == 0
    assert shell.debugger.vm.pc == 4


def test_handle_goto_with_unknown_label(shell, capsys):
    shell.handle_command("goto whatever")

    assert capsys.readouterr().out == "Error: could not locate label `whatever`.\n"


def test_handle_goto_with_too_few_arguments(shell, capsys):
    shell.handle_command("goto")

    assert capsys.readouterr().out == "goto takes one argument.\n"


def test_handle_goto_with_too_many_arguments(shell, capsys):
    shell.handle_command("goto 1 2 3")

    assert capsys.readouterr().out == "goto takes one argument.\n"


def test_handle_goto_abbreviated(shell):
    with patch("hera.debugger.shell.Shell.handle_goto") as mock_handle_goto:
        shell.handle_command("g 10")
        assert mock_handle_goto.call_count == 1

        args, kwargs = mock_handle_goto.call_args
        assert len(args) == 1
        assert args[0] == ["10"]
        assert len(kwargs) == 0


def test_handle_info(shell, capsys):
    shell.handle_command("info")

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
All registers set to zero.

All flags are off.

The call stack is empty.
"""
    )


def test_handle_info_with_registers_and_flags(shell, capsys):
    shell.debugger.vm.registers[7] = 42
    shell.debugger.vm.flag_carry_block = True
    shell.handle_command("info")

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
R7 = 42, all other registers set to zero.

Carry-block flag is on, all other flags are off.

The call stack is empty.
"""
    )


def test_handle_info_with_all_registers_set(shell, capsys):
    shell.debugger.vm.registers = [1] * 16
    shell.handle_command("info")

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
R1 = 1, R2 = 1, R3 = 1, R4 = 1, R5 = 1, R6 = 1, R7 = 1, R8 = 1, R9 = 1, R10 = 1, \
R11 = 1, R12 = 1, R13 = 1, R14 = 1, R15 = 1

All flags are off.

The call stack is empty.
"""
    )


def test_handle_info_with_symbols_arg(shell, capsys):
    shell.debugger.symbol_table["array"] = DataLabel(0xC001)
    shell.handle_command("info symbols")

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
Constants: N (3)
Labels: add (<string>:6)
Data labels: array (0xc001)
"""
    )


def test_handle_info_with_registers_arg(shell, capsys):
    shell.handle_command("info registers")

    assert capsys.readouterr().out == "All registers set to zero.\n"


def test_handle_info_with_flags_arg(shell, capsys):
    shell.handle_command("info flags")

    assert capsys.readouterr().out == "All flags are off.\n"


def test_handle_info_with_multiple_args(shell, capsys):
    shell.handle_command("info symbols flags")

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
Constants: N (3)
Labels: add (<string>:6)
Data labels: array (0xc001)

All flags are off.
"""
    )


def test_handle_info_with_stack_arg(capsys):
    shell = load_shell(
        """\
CALL(FP_alt, do_nothing)
HALT()

LABEL(do_nothing)
  RETURN(FP_alt, PC_ret)
"""
    )

    shell.handle_command("break do_nothing")
    shell.handle_command("continue")
    capsys.readouterr()
    shell.handle_command("info stack")

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
Call stack (last call at bottom)
  do_nothing (<string>:5, called from <string>:1)
"""
    )


def test_handle_info_with_unrecognized_arg(shell, capsys):
    shell.handle_command("info symbols machine")

    captured = capsys.readouterr().out
    assert captured == "Error: unrecognized argument `machine`.\n"


def test_handle_info_abbreviated(shell):
    with patch("hera.debugger.shell.Shell.handle_info") as mock_handle_info:
        shell.handle_command("i")
        assert mock_handle_info.call_count == 1


def test_handle_list(shell, capsys):
    shell.handle_command("list")

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
[<string>]

    1  // A comment
    2  CONSTANT(N, 3)
    3
->  4  SET(R1, N)
    5  SET(R2, 39)
    6  LABEL(add)
    7  ADD(R3, R1, R2)
"""
    )


def test_handle_list_with_context_arg(shell, capsys):
    shell.handle_command("list 1")

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
[<string>]

    3
->  4  SET(R1, N)
    5  SET(R2, 39)
"""
    )


def test_handle_list_with_invalid_context_arg(shell, capsys):
    shell.handle_command("list abc")

    assert capsys.readouterr().out == "Could not parse argument to list.\n"


def test_handle_list_after_end_of_program(capsys):
    shell = load_shell("SET(R1, 42)")
    shell.handle_command("c")
    capsys.readouterr()
    shell.handle_command("list")

    assert capsys.readouterr().out == "Program has finished executing.\n"


def test_handle_list_with_too_many_args(shell, capsys):
    shell.handle_command("list 1 2")

    assert capsys.readouterr().out == "list takes zero or one arguments.\n"


def test_handle_list_abbreviated(shell):
    with patch("hera.debugger.shell.Shell.handle_list") as mock_handle_list:
        shell.handle_command("l")
        assert mock_handle_list.call_count == 1


def test_handle_ll(shell, capsys):
    shell.handle_command("ll")

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
[<string>]

    1  // A comment
    2  CONSTANT(N, 3)
    3
->  4  SET(R1, N)
    5  SET(R2, 39)
    6  LABEL(add)
    7  ADD(R3, R1, R2)
    8  HALT()
    9
"""
    )


def test_handle_ll_with_too_many_args(shell, capsys):
    shell.handle_command("ll 1")

    assert capsys.readouterr().out == "ll takes no arguments.\n"


def test_handle_ll_after_end_of_program(capsys):
    shell = load_shell("SET(R1, 42)")
    shell.handle_command("c")
    capsys.readouterr()
    shell.handle_command("ll")

    assert capsys.readouterr().out == "Program has finished executing.\n"


def test_handle_abbreviated_ll(shell, capsys):
    with patch("hera.debugger.shell.Shell.handle_ll") as mock_handle_ll:
        shell.handle_command("ll")
        assert mock_handle_ll.call_count == 1


def test_handle_on(shell):
    shell.handle_command("on carry-block")
    shell.handle_command("on carry")
    shell.handle_command("on overflow")
    shell.handle_command("on sign")
    shell.handle_command("on zero")

    assert shell.debugger.vm.flag_carry_block
    assert shell.debugger.vm.flag_carry
    assert shell.debugger.vm.flag_overflow
    assert shell.debugger.vm.flag_sign
    assert shell.debugger.vm.flag_zero


def test_handle_on_with_abbreviated_flag(shell):
    shell.handle_command("on cb")
    shell.handle_command("on c")
    shell.handle_command("on v")
    shell.handle_command("on s")
    shell.handle_command("on z")

    assert shell.debugger.vm.flag_carry_block
    assert shell.debugger.vm.flag_carry
    assert shell.debugger.vm.flag_overflow
    assert shell.debugger.vm.flag_sign
    assert shell.debugger.vm.flag_zero


def test_handle_on_with_multiple_args(shell):
    shell.handle_command("on c v")

    assert shell.debugger.vm.flag_carry
    assert shell.debugger.vm.flag_overflow


def test_handle_on_with_invalid_flag(shell, capsys):
    shell.handle_command("on c y")

    assert capsys.readouterr().out == "Unrecognized flag: `y`.\n"


def test_handle_on_with_no_args(shell, capsys):
    shell.handle_command("on")

    assert capsys.readouterr().out == "on takes one or more arguments.\n"


def test_handle_off(shell):
    shell.debugger.vm.flag_carry_block = True
    shell.debugger.vm.flag_carry = True
    shell.debugger.vm.flag_overflow = True
    shell.debugger.vm.flag_sign = True
    shell.debugger.vm.flag_zero = True

    shell.handle_command("off carry-block")
    shell.handle_command("off carry")
    shell.handle_command("off overflow")
    shell.handle_command("off sign")
    shell.handle_command("off zero")

    assert not shell.debugger.vm.flag_carry_block
    assert not shell.debugger.vm.flag_carry
    assert not shell.debugger.vm.flag_overflow
    assert not shell.debugger.vm.flag_sign
    assert not shell.debugger.vm.flag_zero


def test_handle_off_with_abbreviated_flag(shell):
    shell.debugger.vm.flag_carry_block = True
    shell.debugger.vm.flag_carry = True
    shell.debugger.vm.flag_overflow = True
    shell.debugger.vm.flag_sign = True
    shell.debugger.vm.flag_zero = True

    shell.handle_command("off cb")
    shell.handle_command("off c")
    shell.handle_command("off v")
    shell.handle_command("off s")
    shell.handle_command("off z")

    assert not shell.debugger.vm.flag_carry_block
    assert not shell.debugger.vm.flag_carry
    assert not shell.debugger.vm.flag_overflow
    assert not shell.debugger.vm.flag_sign
    assert not shell.debugger.vm.flag_zero


def test_handle_off_with_multiple_args(shell):
    shell.debugger.vm.flag_carry = True
    shell.debugger.vm.flag_overflow = True

    shell.handle_command("off c v")

    assert not shell.debugger.vm.flag_carry
    assert not shell.debugger.vm.flag_overflow


def test_handle_off_with_invalid_flag(shell, capsys):
    shell.handle_command("off c y")

    assert capsys.readouterr().out == "Unrecognized flag: `y`.\n"


def test_handle_off_with_no_args(shell, capsys):
    shell.handle_command("off")

    assert capsys.readouterr().out == "off takes one or more arguments.\n"


def test_handle_restart(shell, capsys):
    shell.handle_command("n")
    shell.handle_command("n")
    capsys.readouterr()

    shell.handle_command("restart")

    vm = shell.debugger.vm
    assert vm.pc == 0
    assert vm.registers[1] == 0
    assert vm.registers[2] == 0
    assert (
        capsys.readouterr().out
        == """\
[<string>]

    3
->  4  SET(R1, N)
    5  SET(R2, 39)
"""
    )


def test_handle_restart_with_too_many_args(shell, capsys):
    shell.handle_command("restart 1")

    assert capsys.readouterr().out == "restart takes no arguments.\n"


def test_restart_cannot_be_abbreviated(shell, capsys):
    with patch("hera.debugger.shell.Shell.handle_restart") as mock_handle_restart:
        shell.handle_command("r")
        assert mock_handle_restart.call_count == 0


def test_handle_assign_to_register(shell):
    shell.handle_command("r12 = 10")

    assert shell.debugger.vm.registers[12] == 10


def test_handle_assign_negative_number_to_register(shell):
    shell.handle_command("r12 = -0xabc")

    assert shell.debugger.vm.registers[12] == -0xABC


def test_handle_assign_to_memory_location(shell):
    shell.debugger.vm.registers[9] = 1000

    shell.handle_command("@R9 = 4000")

    assert shell.debugger.vm.memory[1000] == 4000


def test_handle_assign_to_PC(shell):
    shell.handle_command("pc = 10")

    assert shell.debugger.vm.pc == 10


def test_handle_assign_to_symbol(shell, capsys):
    shell.handle_command("f_c = 10")

    assert capsys.readouterr().out == "Eval error: cannot assign to symbol.\n"
    assert "f_c" not in shell.debugger.symbol_table


def test_handle_assign_to_arithmetic_expression(shell, capsys):
    shell.handle_command("1 + 1 = 3")

    assert (
        capsys.readouterr().out
        == "Eval error: cannot assign to arithmetic expression.\n"
    )


def test_handle_assign_with_undefined_symbol(shell, capsys):
    shell.debugger.vm.registers[4] = 42

    shell.handle_command("r4 = whatever")

    assert shell.debugger.vm.registers[4] == 42
    assert capsys.readouterr().out == "Eval error: whatever is not defined.\n"


def test_handle_assign_register_to_symbol(shell):
    shell.handle_command("r7 = add")

    assert shell.debugger.vm.registers[7] == 4


def test_handle_assign_with_invalid_syntax(shell, capsys):
    shell.handle_command("@ = R5")

    assert capsys.readouterr().out == "Parse error: premature end of input.\n"


def test_handle_assign_with_explicit_command(shell):
    shell.handle_command("assign r7 add")

    assert shell.debugger.vm.registers[7] == 4


def test_handle_assign_with_too_many_args(shell, capsys):
    shell.handle_command("assign r1 r2 r3")

    assert capsys.readouterr().out == "assign takes two arguments.\n"


def test_handle_print_register(shell, capsys):
    shell.handle_command("print R1")

    assert capsys.readouterr().out == "0\n"


def test_handle_print_PC_ret(shell, capsys):
    shell.debugger.vm.registers[13] = 2
    shell.handle_command("print PC_ret")

    assert capsys.readouterr().out == "2 [<string>:5]\n"


def test_handle_print_PC_ret_with_explicit_format(shell, capsys):
    shell.debugger.vm.registers[13] = 2
    shell.handle_command("print :xd PC_ret")

    assert capsys.readouterr().out == "0x0002 = 2\n"


def test_handle_print_with_format_string(shell, capsys):
    shell.debugger.vm.registers[5] = 2
    shell.handle_command("print :bl r5")

    assert capsys.readouterr().out == "0b0000000000000010 [<string>:5]\n"


def test_handle_print_zero_register_with_char_format_string(shell, capsys):
    shell.handle_command("print :c R0")

    assert capsys.readouterr().out == "'\\x00'\n"


def test_handle_print_large_int_with_char_format_string(shell, capsys):
    shell.debugger.vm.registers[5] = 1000
    shell.handle_command("print :c R5")

    assert capsys.readouterr().out == "not an ASCII character\n"


def test_handle_print_unsigned_integer_with_signed_format_string(shell, capsys):
    shell.debugger.vm.registers[5] = 42
    shell.handle_command("print :s R5")

    assert capsys.readouterr().out == "not a signed integer\n"


def test_handle_print_with_restrictive_format_string(shell, capsys):
    shell.debugger.vm.registers[13] = 2
    shell.handle_command("print :o r13")

    assert capsys.readouterr().out == "0o00000002\n"


def test_handle_print_memory_expression(shell, capsys):
    shell.debugger.vm.registers[1] = 4
    shell.debugger.vm.memory[4] = 42

    shell.handle_command("print @r1")

    assert capsys.readouterr().out == "42 = '*'\n"


def test_handle_print_PC(shell, capsys):
    shell.debugger.vm.pc = 4

    shell.handle_command("print PC")

    assert capsys.readouterr().out == "4 [<string>:7]\n"


def test_handle_print_PC_with_nonsense_value(shell, capsys):
    shell.debugger.vm.pc = 300

    shell.handle_command("print pc")

    assert capsys.readouterr().out == "300\n"


def test_handle_print_int(shell, capsys):
    shell.handle_command("print 17")

    assert capsys.readouterr().out == "17\n"


def test_handle_print_arithmetic_expression(shell, capsys):
    shell.handle_command("print :x 21 * (1+1)")

    assert capsys.readouterr().out == "0x002a\n"


def test_handle_print_with_another_arithmetic_expression(shell, capsys):
    shell.debugger.vm.registers[1] = 60

    shell.handle_command("print :d r1-12")

    assert capsys.readouterr().out == "48\n"


def test_handle_print_with_multiple_arguments(shell, capsys):
    shell.debugger.vm.registers[1] = 5
    shell.debugger.vm.registers[2] = 7
    shell.handle_command("print :d r1, r2")

    assert capsys.readouterr().out == "R1 = 5\nR2 = 7\n"


def test_handle_print_symbol(shell, capsys):
    shell.handle_command("print add")

    assert capsys.readouterr().out == "4 [<string>:7]\n"


def test_handle_print_undefined_symbol(shell, capsys):
    shell.handle_command("print whatever")

    assert capsys.readouterr().out == "Eval error: whatever is not defined.\n"


def test_handle_print_invalid_register(shell, capsys):
    shell.handle_command("print R17")

    assert capsys.readouterr().out == "Parse error: R17 is not a valid register.\n"


def test_handle_print_with_division_by_zero(shell, capsys):
    shell.handle_command("print 10 / 0")

    assert capsys.readouterr().out == "Eval error: division by zero.\n"


def test_handle_print_with_nested_division_by_zero(shell, capsys):
    shell.handle_command("print @(10 / 0)")

    assert capsys.readouterr().out == "Eval error: division by zero.\n"


def test_handle_print_with_integer_literal_too_big(shell, capsys):
    shell.handle_command("print 100000")

    assert capsys.readouterr().out == "Eval error: integer literal exceeds 16 bits.\n"


def test_handle_print_with_overflow_from_multiplication(shell, capsys):
    shell.handle_command("print 30000*40")

    assert capsys.readouterr().out == "Eval error: overflow from *.\n"


def test_handle_print_with_integer_literal_too_small(shell, capsys):
    shell.handle_command("print -65000")

    assert capsys.readouterr().out == "Eval error: overflow from unary -.\n"


def test_handle_print_with_invalid_format(shell, capsys):
    shell.handle_command("print :y R1")

    assert capsys.readouterr().out == "Unknown format specifier `y`.\n"


def test_handle_print_undefined_symbol_in_memory_expression(shell, capsys):
    shell.handle_command("print @whatever")

    assert capsys.readouterr().out == "Eval error: whatever is not defined.\n"


def test_handle_print_case_sensitive_symbol(shell, capsys):
    shell.debugger.symbol_table["ADD"] = 10

    shell.handle_command("print ADD")

    assert capsys.readouterr().out == "10\n"


def test_handle_print_with_too_few_args(shell, capsys):
    shell.handle_command("print")

    assert capsys.readouterr().out == "print takes one or more arguments.\n"


def test_handle_print_abbreviated(shell):
    with patch("hera.debugger.shell.Shell.handle_print") as mock_handle_print:
        shell.handle_command("p @R7")
        assert mock_handle_print.call_count == 1

        args, kwargs = mock_handle_print.call_args
        assert len(args) == 1
        assert args[0] == "@R7"
        assert len(kwargs) == 0


def test_handle_doc(shell, capsys):
    shell.handle_command("doc")

    out = capsys.readouterr().out
    assert "SET" in out


def test_handle_doc_with_arguments(shell, capsys):
    shell.handle_command("doc ADD xor Buler")

    out = capsys.readouterr().out
    assert "ADD" in out
    assert "XOR" in out
    assert "BULER" in out


def test_handle_doc_with_nonexistent_operation(shell, capsys):
    shell.handle_command("doc div")

    out = capsys.readouterr().out
    assert "DIV is not a HERA operation" in out


def test_handle_doc_with_BRANCH(shell, capsys):
    shell.handle_command("doc branch")

    out = capsys.readouterr().out
    assert "Register branch" in out
    assert "Relative branch" in out


def test_handle_help(shell, capsys):
    shell.handle_command("help")

    out = capsys.readouterr().out
    assert "Available commands" in out
    assert "Error:" not in out


def test_handle_help_with_one_arg(shell, capsys):
    shell.handle_command("help next")

    out = capsys.readouterr().out
    # Make sure it's not just printing the regular help message.
    assert "Available commands" not in out
    assert "next" in out
    assert "Execute the current line" in out


def test_handle_help_with_abbreviated_command_name(shell, capsys):
    shell.handle_command("help n")

    out = capsys.readouterr().out
    # Make sure it's not just printing the regular help message.
    assert "Available commands" not in out
    assert "next" in out
    assert "Execute the current line" in out


def test_handle_help_with_multiple_args(shell, capsys):
    shell.handle_command("help break next")

    out = capsys.readouterr().out
    # Make sure it's not just printing the regular help message.
    assert "Available commands" not in out
    assert "next" in out
    assert "Execute the current line" in out
    assert "break" in out


def test_handle_help_with_all_commands(shell, capsys):
    shell.handle_command(
        "help assign break clear continue execute help info list ll next off on print \
         restart goto step undo quit asm dis"
    )

    assert "not a recognized command" not in capsys.readouterr().out


def test_handle_help_with_unknown_command(shell, capsys):
    shell.handle_command("help whatever")

    assert capsys.readouterr().out == "whatever is not a recognized command.\n"


def test_handle_help_abbreviated(shell):
    with patch("hera.debugger.shell.Shell.handle_help") as mock_handle_help:
        shell.handle_command("h break")
        assert mock_handle_help.call_count == 1

        args, kwargs = mock_handle_help.call_args
        assert len(args) == 1
        assert args[0] == ["break"]
        assert len(kwargs) == 0


def test_handle_step(capsys):
    shell = load_shell(
        """\
SET(R1, 4)
CALL(FP_alt, plus_two)
SET(R2, 5)
HALT()

LABEL(plus_two)
  INC(R1, 2)
  RETURN(FP_alt, PC_ret)
"""
    )
    shell.handle_command("n")
    capsys.readouterr()
    shell.handle_command("step")

    assert shell.debugger.vm.pc == 8
    assert shell.debugger.vm.registers[1] == 4
    assert shell.debugger.vm.registers[2] == 0

    captured = capsys.readouterr().out
    assert (
        captured
        == """\
[<string>]

    6  LABEL(plus_two)
->  7    INC(R1, 2)
    8    RETURN(FP_alt, PC_ret)
"""
    )


def test_handle_step_not_on_CALL(shell, capsys):
    shell.handle_command("step")

    assert (
        capsys.readouterr().out
        == "step is only valid when the current instruction is CALL.\n"
    )


def test_handle_step_with_too_many_args(shell, capsys):
    shell.handle_command("step 1")

    assert capsys.readouterr().out == "step takes no arguments.\n"


def test_handle_step_abbreviated(shell):
    with patch("hera.debugger.shell.Shell.handle_step") as mock_handle_step:
        shell.handle_command("s")
        assert mock_handle_step.call_count == 1


def test_handle_undo_after_next(shell, capsys):
    shell.handle_command("n")
    capsys.readouterr()
    shell.handle_command("undo")

    assert capsys.readouterr().out == "Undid next.\n"
    assert shell.debugger.vm.pc == 0
    assert shell.debugger.vm.registers[1] == 0


def test_handle_undo_after_next_and_print(shell, capsys):
    shell.handle_command("n")
    shell.handle_command("p r1")
    capsys.readouterr()
    shell.handle_command("undo")

    assert capsys.readouterr().out == "Undid next.\n"
    assert shell.debugger.vm.pc == 0
    assert shell.debugger.vm.registers[1] == 0


def test_handle_undo_after_break_and_continue(shell, capsys):
    shell.handle_command("b add")
    shell.handle_command("continue")
    capsys.readouterr()
    shell.handle_command("undo")
    shell.handle_command("undo")

    assert capsys.readouterr().out == "Undid continue.\nUndid break.\n"
    assert shell.debugger.vm.pc == 0
    assert shell.debugger.vm.registers[1] == 0
    assert shell.debugger.vm.registers[2] == 0
    assert len(shell.debugger.breakpoints) == 0


def test_handle_undo_after_print(shell, capsys):
    shell.handle_command("print r1")
    capsys.readouterr()
    shell.handle_command("undo")

    assert capsys.readouterr().out == "Nothing to undo.\n"


def test_handle_undo_after_nothing(shell, capsys):
    shell.handle_command("undo")

    assert capsys.readouterr().out == "Nothing to undo.\n"


def test_handle_undo_after_assign(shell, capsys):
    shell.handle_command("r1 = 666")
    shell.handle_command("undo")

    assert shell.debugger.vm.registers[1] == 0
    assert capsys.readouterr().out == "Undid assign.\n"


def test_handle_undo_after_clear(shell, capsys):
    shell.handle_command("break 4")
    shell.handle_command("clear *")
    capsys.readouterr()

    shell.handle_command("undo")

    assert len(shell.debugger.breakpoints) == 1
    assert capsys.readouterr().out == "Undid clear.\n"


def test_handle_undo_twice(shell, capsys):
    shell.handle_command("n")
    shell.handle_command("n")
    capsys.readouterr()
    shell.handle_command("undo")
    shell.handle_command("undo")

    assert capsys.readouterr().out == "Undid next.\nUndid next.\n"
    assert shell.debugger.vm.pc == 0
    assert shell.debugger.vm.registers[1] == 0
    assert shell.debugger.vm.registers[2] == 0


def test_handle_undo_with_too_many_args(shell, capsys):
    shell.handle_command("undo 1")

    assert capsys.readouterr().out == "undo takes no arguments.\n"


def test_handle_undo_abbreviated(shell):
    with patch("hera.debugger.shell.Shell.handle_undo") as mock_handle_undo:
        shell.handle_command("u")
        assert mock_handle_undo.call_count == 1


def test_handle_dis(shell, capsys):
    shell.handle_command("dis 0xe1ff")

    assert capsys.readouterr().out == "SETLO(R1, 255)\n"


def test_handle_dis_with_multiple_arguments(shell, capsys):
    shell.handle_command("dis 0xe198 0xf1b7")

    assert capsys.readouterr().out == "SETLO(R1, 152)\nSETHI(R1, 183)\n"


def test_handle_dis_invalid_argument(shell, capsys):
    shell.handle_command("dis abc")

    assert capsys.readouterr().out == "Could not parse argument `abc` to dis.\n"


def test_handle_dis_with_an_OPCODE(capsys):
    shell = load_shell("OPCODE(0x3132)")
    shell.handle_command("dis")

    assert capsys.readouterr().out == "LSR8(R1, R2)\n"


def test_handle_dis_without_an_OPCODE(shell, capsys):
    shell.handle_command("dis")

    assert capsys.readouterr().out == "Current operation is not an OPCODE.\n"

    # Make sure it doesn't choke at the end of program.
    shell.handle_command("continue")
    capsys.readouterr()
    shell.handle_command("dis")

    assert capsys.readouterr().out == "Current operation is not an OPCODE.\n"


def test_handle_asm(shell, capsys):
    shell.handle_command("asm SETLO(R1, 255)")

    assert capsys.readouterr().out == "e1ff\n"


def test_handle_asm_with_only_data(shell, capsys):
    shell.handle_command("asm INTEGER(42)")

    assert capsys.readouterr().out == "49152*0\nc002\n2a\n"


def test_handle_asm_with_data_and_code(shell, capsys):
    shell.handle_command("asm DLABEL(x) INTEGER(42) SET(R1, x)")

    assert (
        capsys.readouterr().out
        == """\
[DATA]
  49152*0
  c002
  2a
[CODE]
  e101
  f1c0
"""
    )


def test_handle_asm_with_too_few_args(shell, capsys):
    shell.handle_command("asm")

    assert capsys.readouterr().out == "asm takes one argument.\n"


def test_handle_unknown_command(shell, capsys):
    shell.handle_command("whatever")

    assert capsys.readouterr().out == "whatever is not a recognized command.\n"


def test_handle_unknown_command_with_arg(shell, capsys):
    shell.handle_command("run program")

    assert capsys.readouterr().out == "run is not a recognized command.\n"


def test_handle_attempt_to_print_register(shell, capsys):
    shell.handle_command("r1")

    assert capsys.readouterr().out == "r1 is not a recognized command.\n"


def test_data_statements(capsys):
    shell = load_shell("DLABEL(X)\nINTEGER(42)\nSET(R1, X)")

    assert shell.debugger.vm.load_memory(DEFAULT_DATA_START) == 42


def test_location_to_instruction_number_with_line_number(debugger):
    assert debugger.location_to_instruction_number("4") == 0


def test_location_to_instruction_number_with_label(debugger):
    assert debugger.location_to_instruction_number("add") == 4


def test_location_to_instruction_number_fails_with_constant(debugger):
    with pytest.raises(ValueError) as e:
        debugger.location_to_instruction_number("N")
    assert "could not locate label `N`" in str(e.value)


def test_location_to_instruction_number_out_of_range(debugger):
    with pytest.raises(ValueError) as e:
        debugger.location_to_instruction_number("100")
    assert "could not find corresponding line" in str(e.value)


def test_location_to_instruction_number_invalid_format(debugger):
    with pytest.raises(ValueError) as e:
        debugger.location_to_instruction_number("a")
    assert "could not locate label `a`" in str(e.value)


def test_instruction_number_to_location(debugger):
    # Zero'th instruction corresponds to second line.
    assert debugger.instruction_number_to_location(0) == "<string>:4"


def test_instruction_number_to_location_with_label(debugger):
    assert debugger.instruction_number_to_location(4) == "<string>:7 (add)"


def test_instruction_number_to_location_does_not_include_constant(debugger):
    assert debugger.instruction_number_to_location(3) == "<string>:5"


def test_print_current_op(shell, capsys):
    shell.print_current_op()

    captured = capsys.readouterr()
    assert (
        captured.out
        == """\
[<string>]

    3
->  4  SET(R1, N)
    5  SET(R2, 39)
"""
    )


def test_reverse_lookup_label():
    symbol_table = {"n": Constant(11), "end": Label(11)}
    assert reverse_lookup_label(symbol_table, 11) == "end"
    assert reverse_lookup_label(symbol_table, 12) is None


def test_debug_empty_program(capsys):
    debug(Program([], [], {}, None), Settings())

    assert capsys.readouterr().out == "Cannot debug an empty program.\n"


def test_label_on_last_line(capsys):
    shell = load_shell("NOP()\nLABEL(my_label)")
    shell.handle_command("info sym")

    captured = capsys.readouterr()
    assert captured.err == ""
    assert "my_label" in captured.out



"""
The definition of all the operations in the HERA language.

Each operation is defined as a class which inherits from `AbstractOperation` or one of
its subclasses. A class for a HERA op typically needs to define only three things:

    - A class-level `P` field that is a tuple of the types of the operations arguments.
      The types are defined in the TYPES FOR `P` FIELD section of this module.

    - A class-level `BITV` field that is a string describing the pattern of the
      operation's binary encoding. See the docstring of `match_bitvector` for a
      description of the pattern format.

    - An `execute` method that takes a `VirtualMachine` object and performs its
      operation on it.

Many subclasses of `AbstractOperation` are provided to ease implementation further, such
as `UnaryOp`, `BinaryOp`, `RegisterBranch`, and `RelativeBranch`.

A class for a HERA pseudo-op typically needs to define only two things:

    - A `P` field, same as regular ops.
    - A `convert` method which converts the pseudo-op into a list of regular operations.

Data operations (e.g., `INTEGER`) should inherit from `DataOperation` and define a `P`
field and `execute` and `assemble` methods. The `assemble` method of a data operation
returns the data, as a `bytes` object, that the operation places into static memory.

Once the operation's class has been defined, make an entry in the `name_to_class`
dictonary in this module. After doing this, the operation should work throughout the
hera-py toolkit (interpreter, debugger, assembler, etc.)!

Author:  Ian Fisher (iafisher@fastmail.com)
Version: July 2019
"""
import json
import sys
from contextlib import suppress

from hera import stdlib
from hera.data import Constant, DataLabel, HERAError, Label, Location, Messages, Token
from hera.utils import format_int, from_u16, print_error, print_warning, to_u16, to_u32
from hera.vm import VirtualMachine


class AbstractOperation:
    """
    The abstract base class for HERA operations. All operation classes should inherit
    from this class.
    """

    # Default value supplied to silence mypy's complaints.
    BITV = ""

    def __init__(self, *args, loc=None):
        self.args = [a.value for a in args]
        self.tokens = list(args)
        # When the preprocessor converts pseudo-ops to real ops, it will set the
        # original field of the real op to the corresponding pseudo-op, for use by the
        # HERA debugger.
        self.original = None

        if isinstance(loc, Location):
            self.loc = loc
        elif hasattr(loc, "location"):
            self.loc = loc.location
        else:
            self.loc = None

    def typecheck(self, symbol_table: "Dict[str, int]") -> Messages:
        """
        Type-check the operation. Subclasses do not generally need to override this
        method, as long as they provide a P class field listing their parameter types.
        """
        messages = Messages()
        if len(self.P) < len(self.tokens):
            msg = "too many args to {} (expected {})".format(self.name, len(self.P))
            messages.err(msg, self.loc)
        elif len(self.P) > len(self.tokens):
            msg = "too few args to {} (expected {})".format(self.name, len(self.P))
            messages.err(msg, self.loc)

        return messages.extend(check_arglist(self.P, self.tokens, symbol_table))

    def convert(self) -> "List[AbstractOperation]":
        """
        Convert the pseudo-operation into a list of real operations. Only pseudo-ops
        need to override this method.
        """
        return [self]

    def assemble(self) -> "Optional[bytes]":
        """
        Assemble the operation into a 16-bit string. Subclasses do not generally need
        to override this method, as long as they provide a BITV class field.
        """
        return substitute_bitvector(self.BITV, self.args)

    @classmethod
    def disassemble(cls, *args):
        """
        Disassemble the integer arguments into an operation. Subclasses do not
        generally need to override this method; it is provided only for the special
        purposes of the INC and DEC classes.
        """
        return cls(*args)

    def execute(self, vm: VirtualMachine) -> None:
        """
        Execute the operation on a virtual machine. Real operations should override
        this method; pseudo-operations should not.
        """
        raise NotImplementedError

    def __getattr__(self, name):
        if name == "name":
            return self.__class__.__name__
        else:
            raise AttributeError(name)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and len(self.tokens) == len(other.tokens)
            and all(a1 == a2 for a1, a2 in zip(self.tokens, other.tokens))
        )

    def __repr__(self):
        return "{}({})".format(self.name, ", ".join(repr(a) for a in self.tokens))

    def __str__(self):
        return "{}({})".format(
            self.name, ", ".join(arg_to_string(a) for a in self.tokens)
        )


def arg_to_string(arg):
    if arg.type == Token.STRING:
        return json.dumps(arg.value)
    elif arg.type == Token.REGISTER:
        return "R" + str(arg.value)
    else:
        return str(arg.value)


# == TYPES FOR `P` FIELD ==
# Use these variables in the `P` field of classes for HERA ops.
REGISTER = "REGISTER"
REGISTER_OR_LABEL = "REGISTER_OR_LABEL"
STRING = "STRING"
LABEL_TYPE = "LABEL_TYPE"
I16 = range(-(2 ** 15), 2 ** 16)
I16_OR_LABEL = "I16_OR_LABEL"
U16 = range(2 ** 16)
I8 = range(-(2 ** 7), 2 ** 8)
I8_OR_LABEL = "I8_OR_LABEL"
U5 = range(2 ** 5)
U4 = range(2 ** 4)


class UnaryOp(AbstractOperation):
    """
    Abstract class to simplify implementation of unary operations. Child classes only
    need to implement the calculate method and set the BITV field.
    """

    P = (REGISTER, REGISTER)

    def execute(self, vm):
        arg = vm.load_register(self.args[1])
        result = self.calculate(vm, arg)
        vm.set_zero_and_sign(result)
        vm.store_register(self.args[0], result)
        vm.pc += 1

    @staticmethod
    def calculate(vm, arg):
        """
        Calculate the result of the unary operation, given a concrete argument.

        The virtual machine is passed in so that the overflow and carry flags can be
        set if necessary.
        """
        raise NotImplementedError


class BinaryOp(AbstractOperation):
    """
    Abstract class to simplify implementation of binary operations. Child classes only
    need to implement the calculate method and set the BITV field.
    """

    P = (REGISTER, REGISTER, REGISTER)

    def execute(self, vm):
        left = vm.load_register(self.args[1])
        right = vm.load_register(self.args[2])
        result = self.calculate(vm, left, right)
        vm.set_zero_and_sign(result)
        vm.store_register(self.args[0], result)
        vm.pc += 1

    @staticmethod
    def calculate(vm, left, right):
        """
        Calculate the result of the binary operation, given the concrete left and right
        arguments.

        The virtual machine is passed in so that the overflow and carry flags can be
        set if necessary.
        """
        raise NotImplementedError


class Branch(AbstractOperation):
    pass


class RegisterBranch(Branch):
    """
    Abstract class to simplify implementation of register branches. Child classes only
    need to implement the should method and set the BITV field.
    """

    P = (REGISTER_OR_LABEL,)

    def execute(self, vm):
        if self.should(vm):
            vm.pc = vm.load_register(self.args[0])
        else:
            vm.pc += 1

    @staticmethod
    def should(vm):
        """
        Return True if branching should occur, based on the virtual machine's state.
        """
        raise NotImplementedError

    def convert(self):
        if self.tokens[0].type == Token.REGISTER:
            # When the argument to the branch is a concrete register.
            return super().convert()
        else:
            # When the argument to the branch is a label, which has already been
            # substituted for its value.
            lbl = self.args[0]
            return [
                SETLO(Token.R(11), Token.Int(lbl & 0xFF)),
                SETHI(Token.R(11), Token.Int(lbl >> 8)),
                self.__class__(Token.R(11)),
            ]


class RelativeBranch(Branch):
    """
    Abstract class to simplify implementation of relative branches. Child classes only
    need to implement the should method and set the BITV field.
    """

    P = (I8_OR_LABEL,)

    def execute(self, vm):
        if self.should(vm):
            vm.pc += self.args[0]
        else:
            vm.pc += 1

    @staticmethod
    def should(vm):
        """
        Return True if branching should occur, based on the virtual machine's state.
        """
        raise NotImplementedError


class DebuggingOperation(AbstractOperation):
    def assemble(self):
        return None


class DataOperation(AbstractOperation):
    pass


class SETLO(AbstractOperation):
    """
    SETLO(Rd, v)
      Set register Rd to the sign-extended 8-bit integer v. Sign-extension means that
      Rd will contain the value of v, interpreted as a signed integer, regardless of
      what was in Rd before. In other words, although the operation is called SETLO, it
      affects not just the low 8 bits but also the high 8 bits.

      Note that values of v above 127 are interpreted as negative numbers, so that for
      instance SETLO(R1, 200) results in R1 having a value of -56 (= 200 - 256).

      SETLO does not affect any flags.

      HERA programmers typically use the SET pseudo-operation to assign a value to a
      register, rather than the low-level SETLO operation.
    """

    P = (REGISTER, I8)
    BITV = "1110 AAAA bbbbbbbb"

    def execute(self, vm):
        value = self.args[1]
        if value > 127:
            value -= 256

        vm.store_register(self.args[0], to_u16(value))
        vm.pc += 1


class SETHI(AbstractOperation):
    """
    SETHI(Rd, v)
      Set the high 8 bits of the register Rd to the unsigned 8-bit integer v. SETHI
      is typically used just after a SETLO operation on the same register.

      SETHI does not affect any flags.

      HERA programmers typically use the SET pseudo-operation to assign a value to a
      register, rather than the low-level SETHI operation.
    """

    P = (REGISTER, I8)
    BITV = "1111 AAAA bbbbbbbb"

    def execute(self, vm):
        target, value = self.args
        vm.store_register(target, (value << 8) + (vm.load_register(target) & 0x00FF))
        vm.pc += 1


class SET(AbstractOperation):
    """
    SET(Rd, v)
      Set the register Rd to the signed 16-bit integer v.

      SET cannot be used to set a register to the value of another register; use MOVE
      for that instead.

      SET does not affect any flags.
    """

    P = (REGISTER, I16_OR_LABEL)

    def convert(self):
        dest = self.tokens[0]
        value = to_u16(self.args[1])
        lo = value & 0xFF
        hi = value >> 8
        return [
            SETLO(dest, Token.Int(lo), loc=self.loc),
            SETHI(dest, Token.Int(hi), loc=self.loc),
        ]


class ADD(BinaryOp):
    """
    ADD(Rd, Ra, Rb)
      Compute Ra + Rb and store the result in Rd. If the carry block flag is off, then
      the carry flag will be added to the sum if it is set, i.e. Ra + Rb + 1 will be
      computed instead.

      If the sum exceeds 16 bits, the higher bits are lost. For example, if an ADD
      results in a sum of 0x10067, then 0x0067 will be stored in the destination
      register.

      ADD sets the following flags:
        sign, if the sum was negative
        zero, if the sum was zero
        carry, if the sum exceeded 16 bits
        overflow, if signed integer overflow occurred (i.e., D != A + B under the signed
                  interpretation of the bits of A, B and D)

      HERA programmers will often want to set the carry block flag (using SETCB())
      before performing this operation, to avoid unexpected results.
    """

    BITV = "1010 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        carry = 1 if not vm.flag_carry_block and vm.flag_carry else 0

        result = (left + right + carry) & 0xFFFF

        vm.flag_carry = result < (left + right + carry)
        vm.flag_overflow = from_u16(result) != from_u16(left) + from_u16(right)

        return result


class SUB(BinaryOp):
    """
    SUB(Rd, Ra, Rb)
      Compute Ra - Rb and store the result in Rd. If the carry block flag is off, then
      the carry flag will be subtracted from the difference if it is NOT set, i.e.
      Ra - Rb - 1 will be computed instead.

      SUB sets the following flags:
        sign, if the difference was negative
        zero, if the difference was zero
        carry, if there was no need to borrow from the 2^16's place (i.e., A > B,
               interpreting A and B as unsigned)
        overflow, if signed integer overflow occurred (i.e., D != A - B under the signed
                  interpretation of the bits of A, B and D)

      HERA programmers will often want to set the carry block flag (using SETCB())
      before performing this operation, to avoid unexpected results.
    """

    BITV = "1011 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        borrow = 1 if not vm.flag_carry_block and not vm.flag_carry else 0

        # to_u16 is necessary because although left and right are necessarily
        # uints, left - right - borrow might not be.
        result = to_u16((left - right - borrow) & 0xFFFF)

        vm.flag_carry = left >= (right + borrow)
        vm.flag_overflow = from_u16(result) != from_u16(left) - from_u16(right) - borrow

        return result


class MUL(BinaryOp):
    """
    MUL(Rd, Ra, Rb)
      Compute Ra * Rb and store the result in Rd.

      The behavior of the MUL instruction depends on the values of the flags. If the
      carry-block is on, or if all the flags are off, then it produces the low 16 bits
      of the product. If the carry-block is off and the sign flag is on, then it
      produces the high 16 bits of the product, to facilitate multiplication of large
      numbers. For all other combinations of flags, the behavior of MUL is undefined by
      the HERA specification; in hera-py, MUL will produce the low 16 bits for the other
      combinations of flags.

      MUL sets the following flags:
        sign, if the product was negative
        zero, if the product was zero
        carry, if the product exceeded 16 bits
        carry, if there was no need to borrow from the 2^16's place (i.e., A > B,
               interpreting A and B as unsigned)
        overflow, if signed integer overflow occurs (i.e., D != A * B under the signed
                  interpretation of the bits of A, B and D)

      HERA programmers will often want to set the carry block flag (using SETCB())
      before performing this operation, to avoid unexpected results.
    """

    BITV = "1100 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        if vm.flag_sign and not vm.flag_carry_block:
            # Take the high 16 bits.
            left = to_u32(from_u16(left))
            right = to_u32(from_u16(right))
            result = ((left * right) & 0xFFFF0000) >> 16
        else:
            # Take the low 16 bits.
            result = (left * right) & 0xFFFF

        vm.flag_carry = result < left * right
        vm.flag_overflow = from_u16(result) != from_u16(left) * from_u16(right)

        return result


class AND(BinaryOp):
    """
    AND(Rd, Ra, Rb)
      Compute the bitwise and of Ra and Rb and store the result in Rd. AND's behavior
      does not depend on the flags, but it does set the sign flag when the result is
      negative and the zero flag when the result is zero.
    """

    BITV = "1000 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        return left & right


class OR(BinaryOp):
    """
    OR(Rd, Ra, Rb)
      Compute the bitwise or of Ra and Rb and store the result in Rd. OR's behavior does
      not depend on the flags, but it does set the sign flag when the result is negative
      and the zero flag when the result is zero.
    """

    BITV = "1001 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        return left | right


class XOR(BinaryOp):
    """
    XOR(Rd, Ra, Rb)
      Compute the bitwise xor of Ra and Rb and store the result in Rd. XOR's behavior
      does not depend on the flags, but it does set the sign flag when the result is
      negative and the zero flag when the result is zero.
    """

    BITV = "1101 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        return left ^ right


class INC(AbstractOperation):
    """
    INC(Rd, v)
      Increment the register by v, an integer between 1 and 64 inclusive. INC sets the
      same flags as the equivalent ADD instruction would, but it ignores the carry flag
      when computing its result.
    """

    P = (REGISTER, range(1, 65))
    BITV = "0011 AAAA 10bb bbbb"

    def execute(self, vm):
        target, value = self.args

        original = vm.load_register(target)
        result = (value + original) & 0xFFFF
        vm.store_register(target, result)

        vm.set_zero_and_sign(result)
        vm.flag_overflow = from_u16(result) != from_u16(original) + value
        vm.flag_carry = value + original >= 2 ** 16
        vm.pc += 1

    def assemble(self):
        # The increment value encoded in the instruction is one less than the actual
        # increment, i.e. INC(R1, 1) is assembled as if it were INC(R1, 0) since the
        # latter is illegal.
        bv = super().assemble()
        return bytes([bv[0], bv[1] - 1])

    @classmethod
    def disassemble(cls, arg0, arg1):
        return cls(arg0, Token(Token.INT, arg1.value + 1))


class DEC(AbstractOperation):
    """
    DEC(Rd, v)
      Decrement the register by v, an integer between 1 and 64 inclusive. DEC sets the
      same flags as the equivalent SUB instruction would, but it ignores the carry flag
      when computing its result.
    """

    P = (REGISTER, range(1, 65))
    BITV = "0011 AAAA 11bb bbbb"

    def execute(self, vm):
        target, value = self.args

        original = vm.load_register(target)
        result = to_u16((original - value) & 0xFFFF)
        vm.store_register(target, result)

        vm.set_zero_and_sign(result)
        vm.flag_overflow = from_u16(result) != from_u16(original) - value
        vm.flag_carry = original >= value
        vm.pc += 1

    def assemble(self):
        # The decrement value encoded in the instruction is one less than the actual
        # decrement, i.e. DEC(R1, 1) is assembled as if it were DEC(R1, 0) since the
        # latter is illegal.
        bv = super().assemble()
        return bytes([bv[0], bv[1] - 1])

    @classmethod
    def disassemble(cls, arg0, arg1):
        return cls(arg0, Token(Token.INT, arg1.value + 1))


class LSL(UnaryOp):
    """
    LSL(Rd, Rb)
      Shift the value of Rb one bit to the left, and store the result in Rd. If the
      carry-block is off and the carry is set, then the bit 1 is shifted in on the
      right; otherwise 0 is shifted in. The bit shifted out becomes the carry flag.

      LSL sets the sign flag when the result is negative, and the zero flag when the
      result is zero.
    """

    BITV = "0011 AAAA 0000 BBBB"

    @staticmethod
    def calculate(vm, arg):
        carry = 1 if vm.flag_carry and not vm.flag_carry_block else 0
        result = ((arg << 1) + carry) & 0xFFFF

        vm.flag_carry = arg & 0x8000

        return result


class LSR(UnaryOp):
    """
    LSR(Rd, Rb)
      Shift the value of Rb one bit to the right, and store the result in Rd. If the
      carry-block is off and the carry is set, then the bit 1 is shifted in on the
      left; otherwise 0 is shifted in. The bit shifted out becomes the carry flag.

      LSR sets the sign flag when the result is negative, and the zero flag when the
      result is zero.
    """

    BITV = "0011 AAAA 0001 BBBB"

    @staticmethod
    def calculate(vm, arg):
        carry = 2 ** 15 if vm.flag_carry and not vm.flag_carry_block else 0
        result = (arg >> 1) + carry

        vm.flag_carry = arg % 2 == 1

        return result


class LSL8(UnaryOp):
    """
    LSL8(Rd, Rb)
      Shift the value of Rb eight bits to the left, and store the result in Rd.

      LSL8 sets the sign flag when the result is negative, and the zero flag when the
      result is zero.
    """

    BITV = "0011 AAAA 0010 BBBB"

    @staticmethod
    def calculate(vm, arg):
        return (arg << 8) & 0xFFFF


class LSR8(UnaryOp):
    """
    LSR8(Rd, Rb)
      Shift the value of Rb eight bits to the right, and store the result in Rd.

      LSR8 sets the sign flag when the result is negative, and the zero flag when the
      result is zero.
    """

    BITV = "0011 AAAA 0011 BBBB"

    @staticmethod
    def calculate(vm, arg):
        return arg >> 8


class ASL(UnaryOp):
    """
    ASL(Rd, Rb)
      Shift the value of Rb one bit to the left, and store the result in Rd. If the
      carry-block is off and the carry is set, then the bit 1 is shifted in on the
      right; otherwise 0 is shifted in. The bit shifted out becomes the carry flag.

      ASL sets the sign flag when the result is negative, and the zero flag when the
      result is zero.

      The only difference between ASL and LSL is that ASL will additionally set the
      overflow flag to the same value it would receive after executing ADD(Rd, Rb, Rb).
    """

    BITV = "0011 AAAA 0100 BBBB"

    @staticmethod
    def calculate(vm, arg):
        carry = 1 if vm.flag_carry and not vm.flag_carry_block else 0
        result = ((arg << 1) + carry) & 0xFFFF

        vm.flag_carry = arg & 0x8000
        vm.flag_overflow = arg & 0x8000 and not result & 0x8000

        return result


class ASR(UnaryOp):
    """
    ASR(Rd, Rb)
      Shift the value of Rb one bit to the right, and store the result in Rd. The bit
      shifted out becomes the carry flag.

      ASR sets the sign flag when the result is negative, and the zero flag when the
      result is zero.

      Unlike LSL, LSR and ASL, ASR ignores the carry flag regardless of whether the
      carry-block is set and always shifts in 0 on the left. The result of ASR is always
      one-half of the original signed value, rounded towards negative infinity.
    """

    BITV = "0011 AAAA 0101 BBBB"

    @staticmethod
    def calculate(vm, arg):
        if arg & 0x8000:
            result = arg >> 1 | 0x8000
        else:
            result = arg >> 1

        vm.flag_carry = arg & 0x0001

        return result


class SAVEF(AbstractOperation):
    """
    SAVEF(Rd)
      Save the flags to Rd. The flags are stored in the following bits:

        0: sign
        1: zero
        2: overflow
        3: carry
        4: carry-block

      The higher bits of Rd are set to 0.
    """

    P = (REGISTER,)
    BITV = "0011 AAAA 0111 0000"

    def execute(self, vm):
        value = (
            int(vm.flag_sign)
            + 2 * int(vm.flag_zero)
            + 4 * int(vm.flag_overflow)
            + 8 * int(vm.flag_carry)
            + 16 * int(vm.flag_carry_block)
        )
        vm.store_register(self.args[0], value)
        vm.pc += 1


class RSTRF(AbstractOperation):
    """
    RSTRF(Rd)
      Restore the flags from Rd.

      Normally you should only invoke RSTRF on a register whose contents have previously
      been set by a call to SAVEF.
    """

    P = (REGISTER,)
    BITV = "0011 AAAA 0111 1000"

    def execute(self, vm):
        value = vm.load_register(self.args[0])
        vm.flag_sign = bool(value & 1)
        vm.flag_zero = bool(value & 0b10)
        vm.flag_overflow = bool(value & 0b100)
        vm.flag_carry = bool(value & 0b1000)
        vm.flag_carry_block = bool(value & 0b10000)
        vm.pc += 1


class FON(AbstractOperation):
    """
    FON(v)
      Turn on the flags indicated by the bits of the 5-bit integer v.

        0: sign
        1: zero
        2: overflow
        3: carry
        4: carry-block
    """

    P = (U5,)
    BITV = "0011 000a 0110 aaaa"

    def execute(self, vm):
        value = self.args[0]
        vm.flag_sign = vm.flag_sign or bool(value & 1)
        vm.flag_zero = vm.flag_zero or bool(value & 0b10)
        vm.flag_overflow = vm.flag_overflow or bool(value & 0b100)
        vm.flag_carry = vm.flag_carry or bool(value & 0b1000)
        vm.flag_carry_block = vm.flag_carry_block or bool(value & 0b10000)
        vm.pc += 1


class FOFF(AbstractOperation):
    """
    FOFF(v)
      Turn off the flags indicated by the bits of the 5-bit integer v. See the docs for
      FON for a list of which bits of v correspond to which flags.
    """

    P = (U5,)
    BITV = "0011 100a 0110 aaaa"

    def execute(self, vm):
        value = self.args[0]
        vm.flag_sign = vm.flag_sign and not bool(value & 1)
        vm.flag_zero = vm.flag_zero and not bool(value & 0b10)
        vm.flag_overflow = vm.flag_overflow and not bool(value & 0b100)
        vm.flag_carry = vm.flag_carry and not bool(value & 0b1000)
        vm.flag_carry_block = vm.flag_carry_block and not bool(value & 0b10000)
        vm.pc += 1


class FSET5(AbstractOperation):
    """
    FSET5(v)
      Set the flags according to the bits of the 5-bit integer v. See the docs for FON
      for a list of which bits of v correspond to which flags.
    """

    P = (U5,)
    BITV = "0011 010a 0110 aaaa"

    def execute(self, vm):
        value = self.args[0]
        vm.flag_sign = bool(value & 1)
        vm.flag_zero = bool(value & 0b10)
        vm.flag_overflow = bool(value & 0b100)
        vm.flag_carry = bool(value & 0b1000)
        vm.flag_carry_block = bool(value & 0b10000)
        vm.pc += 1


class FSET4(AbstractOperation):
    """
    FSET4(v)
      Like FSET5, except it does not affect the carry-block flag.
    """

    P = (U4,)
    BITV = "0011 110a 0110 aaaa"

    def execute(self, vm):
        value = self.args[0]
        vm.flag_sign = bool(value & 1)
        vm.flag_zero = bool(value & 0b10)
        vm.flag_overflow = bool(value & 0b100)
        vm.flag_carry = bool(value & 0b1000)
        vm.pc += 1


class LOAD(AbstractOperation):
    """
    LOAD(Rd, o, Rb)
      Load into Rd the value at memory location Rb + o, where o is a 5-bit unsigned
      integer.

      LOAD sets the sign flag when the value loaded into Rd is negative, and the zero
      flag when it is zero.
    """

    P = (REGISTER, U5, REGISTER)
    BITV = "010b AAAA bbbb CCCC"

    def execute(self, vm):
        target, offset, address = self.args

        result = vm.load_memory(vm.load_register(address) + offset)
        vm.set_zero_and_sign(result)
        vm.store_register(target, result)
        vm.pc += 1


class STORE(AbstractOperation):
    """
    STORE(Rd, o, Rb)
      Store the value in Rd at the memory location Rb + o, where o is a 5-bit unsigned
      integer.

      STORE does not set any flags.
    """

    P = (REGISTER, U5, REGISTER)
    BITV = "011b AAAA bbbb CCCC"

    def execute(self, vm):
        source, offset, address = self.args

        vm.store_memory(vm.load_register(address) + offset, vm.load_register(source))
        vm.pc += 1


class BR(RegisterBranch):
    """
    BR(label)
      Jump unconditionally to the given label.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 0000 0000 AAAA"

    @staticmethod
    def should(vm):
        return True


class BRR(RelativeBranch):
    """
    BRR(n)
      Jump forward or backward n instructions.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 0000 aaaa aaaa"

    def execute(self, vm):
        if self.args[0] != 0:
            vm.pc += self.args[0]
        else:
            vm.halted = True


class BL(RegisterBranch):
    """
    BL(label)
      Jump to the given label if either the sign flag or the overflow flag are on, but
      not if both are.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 0010 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_sign ^ vm.flag_overflow


class BLR(RelativeBranch):
    """
    BLR(n)
      Jump forward or backward n instructions if either the sign flag or the overflow
      flag are on, but not if both are.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 0010 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_sign ^ vm.flag_overflow


class BGE(RegisterBranch):
    """
    BGE(label)
      Jump to the given label if the sign flag or the overflow flag are either both on
      or both off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 0011 0000 AAAA"

    @staticmethod
    def should(vm):
        return not (vm.flag_sign ^ vm.flag_overflow)


class BGER(RelativeBranch):
    """
    BGER(n)
      Jump forward or backward n instructions if the sign flag or the overflow flag are
      either both on or both off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 0011 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not (vm.flag_sign ^ vm.flag_overflow)


class BLE(RegisterBranch):
    """
    BLE(label)
      Jump to the given label under the same conditions as BL, and also if the zero flag
      is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 0100 0000 AAAA"

    @staticmethod
    def should(vm):
        return (vm.flag_sign ^ vm.flag_overflow) or vm.flag_zero


class BLER(RelativeBranch):
    """
    BLER(n)
      Jump forward or backward n instructions under the same conditions as BL, and also
      if the zero flag is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 0100 aaaa aaaa"

    @staticmethod
    def should(vm):
        return (vm.flag_sign ^ vm.flag_overflow) or vm.flag_zero


class BG(RegisterBranch):
    """
    BG(label)
      Jump to the given label whenever BLE would not jump.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 0101 0000 AAAA"

    @staticmethod
    def should(vm):
        return not (vm.flag_sign ^ vm.flag_overflow) and not vm.flag_zero


class BGR(RelativeBranch):
    """
    BGR(n)
      Jump forward or backward n instructions whenever BLE would not jump.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 0101 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not (vm.flag_sign ^ vm.flag_overflow) and not vm.flag_zero


class BULE(RegisterBranch):
    """
    BULE(label)
      Jump to the given label if either the carry flag is off or the zero flag is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 0110 0000 AAAA"

    @staticmethod
    def should(vm):
        return not vm.flag_carry or vm.flag_zero


class BULER(RelativeBranch):
    """
    BULER(n)
      Jump forward or backward n instructions if either the carry flag is off or the
      zero flag is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 0110 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not vm.flag_carry or vm.flag_zero


class BUG(RegisterBranch):
    """
    BUG(label)
      Jump to the given label if the carry flag is on and the zero flag is off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 0111 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_carry and not vm.flag_zero


class BUGR(RelativeBranch):
    """
    BUGR(n)
      Jump forward or backward n instructions if the carry flag is on and the zero flag
      is off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 0111 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_carry and not vm.flag_zero


class BZ(RegisterBranch):
    """
    BZ(label)
      Jump to the given label if the zero flag is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 1000 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_zero


class BZR(RelativeBranch):
    """
    BZR(n)
      Jump forward or backward n instructions if the zero flag is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 1000 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_zero


class BNZ(RegisterBranch):
    """
    BNZ(label)
      Jump to the given label if the zero flag is off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 1001 0000 AAAA"

    @staticmethod
    def should(vm):
        return not vm.flag_zero


class BNZR(RelativeBranch):
    """
    BNZR(n)
      Jump forward or backward n instructions if the zero flag is off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 1001 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not vm.flag_zero


class BC(RegisterBranch):
    """
    BC(label)
      Jump to the given label if the carry flag is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 1010 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_carry


class BCR(RelativeBranch):
    """
    BCR(n)
      Jump forward or backward n instructions if the carry flag is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 1010 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_carry


class BNC(RegisterBranch):
    """
    BNC(label)
      Jump to the given label if the carry flag is off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 1011 0000 AAAA"

    @staticmethod
    def should(vm):
        return not vm.flag_carry


class BNCR(RelativeBranch):
    """
    BNCR(n)
      Jump forward or backward n instructions if the carry flag is off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 1011 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not vm.flag_carry


class BS(RegisterBranch):
    """
    BS(label)
      Jump to the given label if the sign flag is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 1100 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_sign


class BSR(RelativeBranch):
    """
    BSR(n)
      Jump forward or backward n instructions if the sign flag is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 1100 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_sign


class BNS(RegisterBranch):
    """
    BNS(label)
      Jump to the given label if the sign flag is off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 1101 0000 AAAA"

    @staticmethod
    def should(vm):
        return not vm.flag_sign


class BNSR(RelativeBranch):
    """
    BNSR(n)
      Jump forward or backward n instructions if the sign flag is off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 1101 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not vm.flag_sign


class BV(RegisterBranch):
    """
    BV(label)
      Jump to the given label if the overflow flag is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 1110 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_overflow


class BVR(RelativeBranch):
    """
    BVR(n)
      Jump forward or backward n instructions if the overflow flag is on.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 1110 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_overflow


class BNV(RegisterBranch):
    """
    BNV(label)
      Jump to the given label if the overflow flag is off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0001 1111 0000 AAAA"

    @staticmethod
    def should(vm):
        return not vm.flag_overflow


class BNVR(RelativeBranch):
    """
    BNVR(n)
      Jump forward or backward n instructions if the overflow flag is off.

      Run `doc branch` for a detailed explanation of branching instructions.
    """

    BITV = "0000 1111 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not vm.flag_overflow


class CALL_AND_RETURN(Branch):
    P = (REGISTER, REGISTER_OR_LABEL)

    def execute(self, vm):
        ra, rb = self.args

        old_pc = vm.pc
        vm.pc = vm.load_register(rb)
        vm.store_register(rb, old_pc + 1)
        # FP = R14
        old_fp = vm.load_register(14)
        vm.store_register(14, vm.load_register(ra))
        vm.store_register(ra, old_fp)

    def typecheck(self, *args, **kwargs):
        messages = super().typecheck(*args, **kwargs)
        if len(self.tokens) >= 1 and self.tokens[0].type == Token.REGISTER:
            with suppress(ValueError):
                if self.tokens[0].value != 12:
                    msg = "first argument to {} should be R12".format(self.name)
                    messages.warn(msg, self.tokens[0])
        return messages


class CALL(CALL_AND_RETURN):
    """
    CALL(FP_alt, function)
      Call the function. The first argument doesn't have to be FP_alt, but you should
      use FP_alt unless you have a good reason not to.

      Unlike in higher-level languages like Python, function calls in HERA don't have
      explicit argument lists. Instead, the caller passes arguments by either placing
      them in designated registers or on designated locations on the stack. The contract
      of where to put arguments is known as the calling convention. See the HERA manual
      for details.

    CALL(Ra, Rb)
      The general-purpose form of the CALL instruction. See the HERA manual for details.
    """

    P = (REGISTER, REGISTER_OR_LABEL)
    BITV = "0010 0000 AAAA BBBB"

    def convert(self):
        if self.tokens[1].type == Token.REGISTER:
            return super().convert()
        else:
            return SET(Token.R(13), self.tokens[1]).convert() + [
                self.__class__(self.tokens[0], Token.R(13))
            ]

    def execute(self, vm):
        # Push a (call_address, return_address) pair onto the debugging call stack.
        vm.expected_returns.append((vm.load_register(self.args[1]), vm.pc + 1))
        super().execute(vm)


class RETURN(CALL_AND_RETURN):
    """
    RETURN(FP_alt, PC_ret)
      Return from a function call. The arguments don't have to be FP_alt and PC_ret, but
      you should use them unless you have a good reason not to.

      Unlike in higher-level languages like Python, function returns in HERA don't have
      explicit return values. Instead, the function returns a value by either placing
      it in a designated register or on a designated location on the stack. See the HERA
      manual for details.

    RETURN(Ra, Rb)
      The general-purpose form of the RETURN instruction. See the HERA manual for
      details.
    """

    P = (REGISTER, REGISTER)
    BITV = "0010 0001 AAAA BBBB"

    def typecheck(self, *args, **kwargs):
        messages = super().typecheck(*args, **kwargs)
        if len(self.tokens) >= 2 and self.tokens[1].type == Token.REGISTER:
            with suppress(ValueError):
                if self.tokens[1].value != 13:
                    messages.warn(
                        "second argument to RETURN should be R13", self.tokens[1]
                    )
        return messages

    def execute(self, vm):
        got = vm.load_register(self.args[1])
        if vm.settings.warn_return_on:
            if vm.expected_returns:
                _, expected = vm.expected_returns.pop()
                if expected != got:
                    msg = "incorrect return address (got {}, expected {})".format(
                        got, expected
                    )
                    print_warning(vm.settings, msg, loc=vm.location)
                    vm.settings.warning_count += 1
            else:
                msg = "incorrect return address (got {}, expected <nothing>)".format(
                    got
                )
                print_warning(vm.settings, msg, loc=vm.location)
                vm.settings.warning_count += 1
        super().execute(vm)


class SWI(AbstractOperation):
    """
    SWI(i)
      Simulate a software interrupt that has the number i, a 4-bit integer. The
      counterpart to this instruction is RTI.
    """

    P = (U4,)
    BITV = "0010 0010 0000 aaaa"


class RTI(AbstractOperation):
    """
    RTI()
      Return from a software interrupt. The counterpart to this instruction is SWI.
    """

    P = ()
    BITV = "0010 0011 0000 0000"


class CMP(AbstractOperation):
    """
    CMP(Ra, Rb)
      Compare the values of Ra and Rb. The branch instructions are named so that they
      do the expected thing when immediately preceded by a CMP instruction, e.g.
      CMP(R1, R2) BLE(somewhere_else) will jump to somewhere_else if R1 <= R2.

      At the machine level, this instruction sets the same flags as would be set for
      subtracting Rb from Ra.
    """

    P = (REGISTER, REGISTER)

    def convert(self):
        return [FON(Token.Int(8)), SUB(Token.R(0), self.tokens[0], self.tokens[1])]


class CON(AbstractOperation):
    """
    CON()
      Turn the carry flag on. See also COFF.
    """

    P = ()

    def convert(self):
        return [FON(Token.Int(8))]


class COFF(AbstractOperation):
    """
    COFF()
      Turn the carry flag off. See also CON.
    """

    P = ()

    def convert(self):
        return [FOFF(Token.Int(8))]


class CBON(AbstractOperation):
    """
    CBON()
      Turn the carry-block flag on. See also CCBOFF.
    """

    P = ()

    def convert(self):
        return [FON(Token.Int(16))]


class CCBOFF(AbstractOperation):
    """
    CCBOFF()
      Turn the carry and carry-block flags off. See also CBON and CON.
    """

    P = ()

    def convert(self):
        return [FOFF(Token.Int(24))]


class MOVE(AbstractOperation):
    """
    MOVE(Ra, Rb)
      Set Ra to the value of Rb. Equivalent to an assignment statement in a higher-level
      language.
    """

    P = (REGISTER, REGISTER)

    def convert(self):
        return [OR(self.tokens[0], self.tokens[1], Token.R(0))]


class SETRF(AbstractOperation):
    """
    SETRF(Rd, v)
      Set Rd to v and set the flags for v + 0.
    """

    P = (REGISTER, I16_OR_LABEL)

    def convert(self):
        return SET(*self.args).convert() + FLAGS(self.tokens[0]).convert()


class FLAGS(AbstractOperation):
    """
    FLAGS(Ra)
      Set the flags for Ra + 0.
    """

    P = (REGISTER,)

    def convert(self):
        return [FOFF(Token.Int(8)), ADD(Token.R(0), self.tokens[0], Token.R(0))]


class HALT(AbstractOperation):
    """
    HALT()
      Stop execution of the program, permanently.
    """

    P = ()

    def convert(self):
        return [BRR(Token.Int(0))]


class NOP(AbstractOperation):
    """
    NOP()
      Do nothing.
    """

    P = ()

    def convert(self):
        return [BRR(Token.Int(1))]


class NEG(AbstractOperation):
    """
    NEG(Rd, Ra)
      Compute the arithmetic negation of Ra and store it in Rd. Flags are set as for
      0 - Ra.
    """

    P = (REGISTER, REGISTER)

    def convert(self):
        return [FON(Token.Int(8)), SUB(self.tokens[0], Token.R(0), self.tokens[1])]


class NOT(AbstractOperation):
    """
    NOT(Rd, Ra)
      Compute the logical bitwise negation of Ra and store it in Rd. The sign flag is
      set if the result is negative, and the zero flag if it is zero; no other flags
      are affected.
    """

    P = (REGISTER, REGISTER)

    def typecheck(self, *args, **kwargs):
        messages = super().typecheck(*args, **kwargs)
        if len(self.tokens) == 2 and self.tokens[1].type == Token.REGISTER:
            with suppress(ValueError):
                if self.tokens[1].value == 11:
                    messages.warn("don't use R11 with NOT", self.tokens[1])
        return messages

    def convert(self):
        return [
            SETLO(Token.R(11), Token.Int(0xFF)),
            SETHI(Token.R(11), Token.Int(0xFF)),
            XOR(self.tokens[0], Token.R(11), self.tokens[1]),
        ]


class OPCODE(AbstractOperation):
    """
    OPCODE(d)
      Interpret the 16-bit integer d as a binary encoding of a HERA instruction, and
      execute that instruction.

      There is no good reason to use this, but if you want you can convert your entire
      program into OPCODE instructions by running

        hera preprocess --obfuscate myprogram.hera
    """

    P = (U16,)

    def typecheck(self, *args, **kwargs):
        messages = super().typecheck(*args, **kwargs)

        try:
            disassemble(self.args[0])
        except HERAError:
            messages.err("not a HERA instruction", self.tokens[0])

        return messages

    def convert(self):
        # We don't need to call .convert() on the return value of disassemble because
        # disassemble never returns a pseudo-op.
        return [disassemble(self.args[0])]


class INTEGER(DataOperation):
    """
    INTEGER(i)
      Place the 16-bit signed integer i into the current data cell.

      INTEGER is a data instruction.
    """

    P = (I16,)

    def execute(self, vm):
        vm.store_memory(vm.dc, to_u16(self.args[0]))
        vm.dc += 1

    def assemble(self):
        return bytes([self.args[0] & 0xFF00, self.args[0] & 0xFF])


class DSKIP(DataOperation):
    """
    DSKIP(n)
      Skip the next n data cells. Typically used to reserve space for a fixed-size
      array.

      DSKIP is a data instruction.
    """

    P = (U16,)

    def execute(self, vm):
        vm.dc += self.args[0]

    def assemble(self):
        return bytes([0] * self.args[0] * 2)


class LP_STRING(DataOperation):
    """
    LP_STRING(s)
      Place the length-prefixed string s into memory at the current data cell, so that
      the first cell holds the string's length and the rest of the cells hold the
      ASCII characters of the string.

      LP_STRING is a data instruction.
    """

    P = (STRING,)

    def execute(self, vm):
        vm.store_memory(vm.dc, len(self.args[0]))
        vm.dc += 1
        for c in self.args[0]:
            vm.store_memory(vm.dc, ord(c))
            vm.dc += 1

    def assemble(self):
        s = self.args[0]
        length_bytes = [len(s) & 0xFF00, len(s) & 0xFF]
        data_bytes = []
        for c in s:
            data_bytes.append(0)
            data_bytes.append(ord(c))
        return bytes(length_bytes + data_bytes)


class CONSTANT(DataOperation):
    """
    CONSTANT(x, d)
      Declare the symbol x to have the value d. x can subsequently be used wherever in
      the program the integer d would have been allowed, e.g. in a SET instruction.
    """

    P = (LABEL_TYPE, I16)

    def convert(self):
        return []


class LABEL(AbstractOperation):
    """
    LABEL(l)
      Designate the next instruction as l. Used so that the next instruction can be the
      target of a branching instruction.
    """

    P = (LABEL_TYPE,)

    def convert(self):
        return []


class DLABEL(DataOperation):
    """
    DLABEL(l)
      Designate the next data cell as l. Used so that instructions like LOAD and STORE
      can refer to data declared in the data segment.

      DLABEL is a data instruction.
    """

    P = (LABEL_TYPE,)

    def convert(self):
        return []


class PRINT_REG(DebuggingOperation):
    """
    print_reg(Ra)
      Print the value of Ra.

      print_reg is a debugging instruction.
    """

    P = (REGISTER,)

    def execute(self, vm):
        v = vm.load_register(self.args[0])
        print("R{} = {}".format(self.args[0], format_int(v)))
        vm.pc += 1


class PRINT(DebuggingOperation):
    """
    print(s)
      Print the string literal, without a newline at the end.

      print is a debugging instruction.
    """

    P = (STRING,)

    def execute(self, vm):
        print(self.args[0], end="")
        vm.pc += 1


class PRINTLN(DebuggingOperation):
    """
    println(s)
      Print the string literal with a newline at the end.

      println is a debugging instruction.
    """

    P = (STRING,)

    def execute(self, vm):
        print(self.args[0])
        vm.pc += 1


class __EVAL(DebuggingOperation):
    P = (STRING,)

    def execute(self, vm):
        try:
            eval(self.args[0], {}, {"stdlib": stdlib, "vm": vm})
        except Exception as e:
            print_error(vm.settings, "Python exception: " + str(e), loc=vm.location)
            sys.exit(3)
        vm.pc += 1


def disassemble(v: int) -> AbstractOperation:
    """Disassemble a 16-bit integer into a HERA operation."""
    # Iterating over every HERA class is inefficient but simple.
    for cls in name_to_class.values():
        if cls.BITV != "":
            m = match_bitvector(cls.BITV, v)
            if m is not False and isinstance(m, list):
                return cls.disassemble(*m)

    raise HERAError("bit pattern does not correspond to HERA instruction")


def match_bitvector(pattern: str, v: int) -> "Union[List, bool]":
    """
    Try to match the 16-bit integer `v` against `pattern`. Return a list of extracted
    arguments if `v` matches, or False otherwise.

    `pattern` should be a string of sixteen characters, which may be the digits '0' or
    '1' or Latin letters. The digits are matched against the literal digits in `v`;
    letters are used to extract values. Uppercase letters match integer values and
    lowercase letters match registers. For example, "0000aaaaBBBB1111" would match,
    e.g., 0000 1001 0110 11111, returning

        [Token(Token.INT, 0b1001), Token(Token.REGISTER, 0b0110)]
    """
    s = bin(v)[2:].rjust(16, "0")
    args = []  # type: List[Token]

    pattern = pattern.replace(" ", "")

    # Initialize argument list with the correct types (integers or registers).
    for pattern_bit in pattern:
        if pattern_bit.isalpha():
            index = ord(pattern_bit.lower()) - ord("a")
            while index >= len(args):
                args.append(Token(Token.INT, 0))

            if pattern_bit.isupper():
                args[index] = Token(Token.REGISTER, 0)
            else:
                args[index] = Token(Token.INT, 0)

    # Fill in the argument list with the actual values from the binary data.
    for pattern_bit, real_bit in zip(pattern, s):
        if pattern_bit == "0":
            if real_bit != "0":
                return False
        elif pattern_bit == "1":
            if real_bit != "1":
                return False
        else:
            index = ord(pattern_bit.lower()) - ord("a")
            args[index].value <<= 1
            if real_bit == "1":
                args[index].value += 1

    return args


def substitute_bitvector(pattern: str, args: "List[int]") -> bytes:
    """
    Given a 16-bit pattern and a list of arguments, substitute the arguments into the
    pattern to yield a two-byte machine operation.
    """
    # Make a copy of the arguments as we will be modifying it.
    args = args[:]

    # Remove spaces from the pattern string.
    pattern = pattern.replace(" ", "")

    # Substituting the low byte MUST happen before the high byte, because it pulls the
    # low bits from each argument.
    low_pattern = pattern[8:16]
    low_byte = substitute_half_a_bitvector(low_pattern, args)

    high_pattern = pattern[0:8]
    high_byte = substitute_half_a_bitvector(high_pattern, args)

    return bytes([high_byte, low_byte])


def substitute_half_a_bitvector(pattern: str, args: "List[int]") -> int:
    """
    Helper function for `substitute_bitvector`. Takes an 8-bit pattern and returns a
    single byte.
    """
    ret = 0
    for shift, pattern_bit in enumerate(reversed(pattern)):
        if pattern_bit == "0":
            real_bit = 0
        elif pattern_bit == "1":
            real_bit = 1
        else:
            index = ord(pattern_bit.lower()) - ord("a")
            real_bit = args[index] & 1
            args[index] >>= 1

        ret += real_bit << shift

    return ret


def check_arglist(argtypes, args, symbol_table):
    """
    Check that the values in `args` match the types in `argtypes`, and return a
    `Messages` object with any warnings or errors generated.
    """
    messages = Messages()
    for expected, got in zip(argtypes, args):
        if expected == REGISTER:
            err = check_register(got)
        elif expected == REGISTER_OR_LABEL:
            err = check_register_or_label(got, symbol_table)
        elif expected == LABEL_TYPE:
            err = check_label(got)
        elif expected == STRING:
            err = check_string(got)
        elif expected == I16_OR_LABEL:
            err = check_in_range(
                got, symbol_table, lo=-(2 ** 15), hi=2 ** 16, labels=True
            )
        elif expected == I8_OR_LABEL:
            err = check_in_range(
                got, symbol_table, lo=-(2 ** 7), hi=2 ** 8, labels=True
            )
        elif isinstance(expected, range):
            err = check_in_range(got, symbol_table, lo=expected.start, hi=expected.stop)
        else:
            raise RuntimeError("unknown parameter type {!r}".format(expected))

        if err is not None:
            messages.err(err, got)
    return messages


def check_register(arg) -> "Optional[str]":
    """
    Check that `arg` is a register. Return an error message as a string if it is not.
    """
    if arg.type == Token.REGISTER:
        return None
    else:
        if isinstance(arg.value, str) and arg.value.lower() == "pc":
            return "program counter cannot be accessed or changed directly"
        else:
            return "expected register"


def check_register_or_label(arg, symbol_table: "Dict[str, int]") -> "Optional[str]":
    """
    Check that `arg` is a register or a label. Return an error message as a string if it
    is not.
    """
    if arg.type == Token.REGISTER:
        return None
    elif arg.type == Token.SYMBOL:
        try:
            val = symbol_table[arg.value]
        except KeyError:
            return "undefined symbol"
        else:
            if isinstance(val, Constant):
                return "constant cannot be used as label"
            elif isinstance(val, DataLabel):
                return "data label cannot be used as branch label"
            else:
                return None
    else:
        return "expected register or label"


def check_label(arg) -> "Optional[str]":
    """Check that `arg` is a label. Return an error message as a string if it is not."""
    if arg.type == Token.SYMBOL:
        return None
    else:
        return "expected label"


def check_string(arg) -> "Optional[str]":
    """
    Check that `arg` is a string literal. Return an error message as a string if it is
    not.
    """
    if not isinstance(arg, Token) or arg.type != Token.STRING:
        return "expected string literal"
    else:
        return None


def check_in_range(arg, symbol_table, *, lo, hi, labels=False) -> "Optional[str]":
    """
    Check that `arg` is an integer (or a symbol resolving to an integer) within the
    bounds established by `lo` and `hi`.

    If `labels` is True, then `arg` may be a label; otherwise it may only be a constant
    symbol.

    Return an error message as a string if `arg` is not an integer or is not in the
    given bounds.
    """
    if arg.type == Token.SYMBOL:
        try:
            arg = Token.Int(symbol_table[arg.value])
        except KeyError:
            return "undefined constant"
        else:
            if not labels and not isinstance(arg.value, Constant):
                return "cannot use label as constant"
            elif labels and isinstance(arg.value, Label):
                return None

    if arg.type != Token.INT:
        return "expected integer"

    if arg.value < lo or arg.value >= hi:
        return "integer must be in range [{}, {})".format(lo, hi)
    else:
        return None


name_to_class = {
    "ADD": ADD,
    "AND": AND,
    "ASL": ASL,
    "ASR": ASR,
    "BC": BC,
    "BCR": BCR,
    "BG": BG,
    "BGR": BGR,
    "BGE": BGE,
    "BGER": BGER,
    "BL": BL,
    "BLR": BLR,
    "BLE": BLE,
    "BLER": BLER,
    "BNC": BNC,
    "BNCR": BNCR,
    "BNS": BNS,
    "BNSR": BNSR,
    "BNV": BNV,
    "BNVR": BNVR,
    "BNZ": BNZ,
    "BNZR": BNZR,
    "BR": BR,
    "BRR": BRR,
    "BS": BS,
    "BSR": BSR,
    "BUG": BUG,
    "BUGR": BUGR,
    "BULE": BULE,
    "BULER": BULER,
    "BV": BV,
    "BVR": BVR,
    "BZ": BZ,
    "BZR": BZR,
    "CALL": CALL,
    "CBON": CBON,
    "CCBOFF": CCBOFF,
    "CMP": CMP,
    "COFF": COFF,
    "CON": CON,
    "CONSTANT": CONSTANT,
    "DEC": DEC,
    "DLABEL": DLABEL,
    "DSKIP": DSKIP,
    "FLAGS": FLAGS,
    "FOFF": FOFF,
    "FON": FON,
    "FSET4": FSET4,
    "FSET5": FSET5,
    "HALT": HALT,
    "INC": INC,
    "INTEGER": INTEGER,
    "LABEL": LABEL,
    "LOAD": LOAD,
    "LP_STRING": LP_STRING,
    "LSL": LSL,
    "LSL8": LSL8,
    "LSR": LSR,
    "LSR8": LSR8,
    "MOVE": MOVE,
    "MUL": MUL,
    "NEG": NEG,
    "NOP": NOP,
    "NOT": NOT,
    "OPCODE": OPCODE,
    "OR": OR,
    "print": PRINT,
    "println": PRINTLN,
    "print_reg": PRINT_REG,
    "RETURN": RETURN,
    "RSTRF": RSTRF,
    "RTI": RTI,
    "SAVEF": SAVEF,
    "SET": SET,
    "SETHI": SETHI,
    "SETLO": SETLO,
    "SETRF": SETRF,
    "STORE": STORE,
    "SUB": SUB,
    "SWI": SWI,
    "TIGER_STRING": LP_STRING,
    "XOR": XOR,
    "__eval": __EVAL,
}



import pytest

from hera.main import main
from .utils import execute_program_helper


def test_error_message_for_missing_comma(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SETLO(R1 40)")

    captured = capsys.readouterr()
    assert "SETLO(R1 40)" in captured.err
    assert "line 1" in captured.err
    assert "col 10" in captured.err


def test_error_message_for_invalid_register(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SET(R17, 65)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: R17 is not a valid register, line 1 col 5 of <stdin>

  SET(R17, 65)
      ^

"""
    )


def test_error_message_for_invalid_register_with_weird_syntax(capsys):
    program = """\
SET(
\tR17,
\t65)
"""
    with pytest.raises(SystemExit):
        execute_program_helper(program)

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: R17 is not a valid register, line 2 col 2 of <stdin>

  \tR17,
  \t^

"""
    )


def test_multiple_error_messages(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("ADD(R1, 10)\nINC(R4)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: too few args to ADD (expected 3), line 1 col 1 of <stdin>

  ADD(R1, 10)
  ^

Error: expected register, line 1 col 9 of <stdin>

  ADD(R1, 10)
          ^

Error: too few args to INC (expected 2), line 2 col 1 of <stdin>

  INC(R4)
  ^

"""
    )


def test_error_message_from_include(capsys):
    with pytest.raises(SystemExit):
        main(["test/assets/error/from_include.hera"])

    captured = capsys.readouterr()
    assert "SET(R1, R2)  // error!" in captured.err
    assert "ADD(R1, R2)  // error!" in captured.err
    assert "test/assets/error/from_include.hera" in captured.err
    assert "test/assets/error/included.hera" in captured.err


def test_dskip_overflow_program(capsys):
    program = """\
DSKIP(0xFFFF)
DLABEL(N)
INTEGER(42)

SET(R1, N)
    """
    with pytest.raises(SystemExit):
        execute_program_helper(program)

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: past the end of available memory, line 1 col 1 of <stdin>

  DSKIP(0xFFFF)
  ^

"""
    )


def test_error_for_use_of_constant_before_declaration(capsys):
    program = "DSKIP(N)\nCONSTANT(N, 100)"

    with pytest.raises(SystemExit):
        execute_program_helper(program)

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: undefined constant, line 1 col 7 of <stdin>

  DSKIP(N)
        ^

"""
    )


def test_error_for_dskip_with_register(capsys):
    program = "DSKIP(R1)"

    with pytest.raises(SystemExit):
        execute_program_helper(program)

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: expected integer, line 1 col 7 of <stdin>

  DSKIP(R1)
        ^

"""
    )


def test_error_for_use_of_constant_before_include(capsys):
    program = 'SET(R1, N)\n#include "test/assets/error/constant_decl.hera"'

    with pytest.raises(SystemExit):
        execute_program_helper(program)

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: undefined constant, line 1 col 9 of <stdin>

  SET(R1, N)
          ^

Error: data statement after code, line 1 col 1 of test/assets/error/constant_decl.hera

  CONSTANT(N, 100)
  ^

"""
    )


def test_error_for_recursive_constant_declaration(capsys):
    program = "CONSTANT(N, N)"

    with pytest.raises(SystemExit):
        execute_program_helper(program)

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: undefined constant, line 1 col 13 of <stdin>

  CONSTANT(N, N)
              ^

"""
    )


def test_constant_redefinition_error(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("CONSTANT(N, 1)\nCONSTANT(N, 2)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: symbol `N` has already been defined, line 2 col 1 of <stdin>

  CONSTANT(N, 2)
  ^

"""
    )


def test_data_after_code(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SET(R1, 1)\nINTEGER(42)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: data statement after code, line 2 col 1 of <stdin>

  INTEGER(42)
  ^

"""
    )


def test_error_message_after_symbol_table_error(capsys):
    # Make sure that an error generating the symbol table doesn't immediately end the
    # program--other type errors should still be caught.
    with pytest.raises(SystemExit):
        execute_program_helper("DSKIP(N)\nADD(R1, R2)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: undefined constant, line 1 col 7 of <stdin>

  DSKIP(N)
        ^

Error: too few args to ADD (expected 3), line 2 col 1 of <stdin>

  ADD(R1, R2)
  ^

"""
    )


def test_overflowing_CONSTANT(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("CONSTANT(N, 0xFFFFF)\nSET(R1, N)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: integer must be in range [-32768, 65536), line 1 col 13 of <stdin>

  CONSTANT(N, 0xFFFFF)
              ^

"""
    )


def test_overflowing_integer_literal(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SET(R1, 0xFFFFF)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: integer must be in range [-32768, 65536), line 1 col 9 of <stdin>

  SET(R1, 0xFFFFF)
          ^

"""
    )


def test_relative_branching_too_far(capsys):
    program = "BRR(l)\n" + ("SET(R1, 1)\n" * 100) + "LABEL(l)"
    with pytest.raises(SystemExit):
        execute_program_helper(program)

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: label is too far for a relative branch, line 1 col 5 of <stdin>

  BRR(l)
      ^

"""
    )


def test_nonexistent_include(capsys):
    program = '#include "unicorn"'
    with pytest.raises(SystemExit):
        execute_program_helper(program)

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: file "unicorn" does not exist, line 1 col 10 of <stdin>

  #include "unicorn"
           ^

"""
    )


def test_invalid_octal_number(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SET(R1, 018)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Warning: consider using "0o" prefix for octal numbers, line 1 col 9 of <stdin>

  SET(R1, 018)
          ^

Error: invalid integer literal, line 1 col 9 of <stdin>

  SET(R1, 018)
          ^

"""
    )


def test_include_before_data(capsys):
    program = """\
#include "test/assets/error/included_before_data.hera"

DLABEL(X)
INTEGER(42)

SET(R1, X)
LOAD(R2, 0, R1)
"""
    with pytest.raises(SystemExit):
        execute_program_helper(program)

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: data statement after code, line 3 col 1 of <stdin>

  DLABEL(X)
  ^

Error: data statement after code, line 4 col 1 of <stdin>

  INTEGER(42)
  ^

"""
    )


def test_error_for_interrupt_instructions(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SWI(10)\nRTI()")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: hera-py does not support SWI, line 1 col 1 of <stdin>

  SWI(10)
  ^

Error: hera-py does not support RTI, line 2 col 1 of <stdin>

  RTI()
  ^

"""
    )


def test_error_for_python_exception_in___eval(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper('__eval("10 / 0")')

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: Python exception: division by zero, line 1 col 1 of <stdin>

  __eval("10 / 0")
  ^

"""
    )


def test_mega_error(capsys):
    with pytest.raises(SystemExit):
        main(["test/assets/error/mega_error.hera"])

    captured = capsys.readouterr().err
    assert (
        captured
        == """\
Warning: consider using "0o" prefix for octal numbers, line 11 col 9 of test/assets/error/mega_error.hera

  SET(R2, 01)
          ^

Warning: unrecognized backslash escape, line 17 col 13 of test/assets/error/mega_error.hera

  LP_STRING("\\y")
              ^

Error: expected comma or right parenthesis, line 2 col 8 of test/assets/error/mega_error.hera

  SET(R1 40)
         ^

Error: over-long character literal, line 14 col 11 of test/assets/error/mega_error.hera

  SETLO(R7, 'ab')
            ^

Error: unclosed string literal, line 20 col 9 of test/assets/error/mega_error.hera

  SET(R1, "
          ^

Error: expected integer, line 5 col 6 of test/assets/error/mega_error.hera

  FOFF(R1)
       ^

Error: too few args to ADD (expected 3), line 8 col 1 of test/assets/error/mega_error.hera

  ADD('c', "abc")
  ^

Error: expected register, line 8 col 5 of test/assets/error/mega_error.hera

  ADD('c', "abc")
      ^

Error: expected register, line 8 col 10 of test/assets/error/mega_error.hera

  ADD('c', "abc")
           ^

Error: data statement after code, line 17 col 1 of test/assets/error/mega_error.hera

  LP_STRING("\\y")
  ^

"""
    )



"""Test programs generated by a compiler for the Tiger language implemented for CS350.
"""
import json
import pytest
from io import StringIO
from unittest.mock import patch

from .utils import execute_program_helper

from hera.main import main
from hera.utils import from_u16


def test_div_and_print(capsys):
    main(["test/assets/cs350/div_and_print.hera"])

    captured = capsys.readouterr()
    assert captured.out == "5"


def test_merge_sort(capsys):
    main(["test/assets/cs350/merge_sort.hera"])

    captured = capsys.readouterr()
    assert captured.out == "42"


def test_lexical_scope_deep(capsys):
    main(["test/assets/cs350/lexical_scope_deep.hera"])

    captured = capsys.readouterr()
    assert captured.out == "42"


def test_array_madness(capsys):
    main(["test/assets/cs350/array_madness.hera"])

    captured = capsys.readouterr()
    assert captured.out == "42"


def test_record_trees(capsys):
    main(["test/assets/cs350/record_trees.hera"])

    captured = capsys.readouterr()
    assert captured.out == "42"


def test_stdlib_ord():
    vm = execute_program_helper(helper_stdlib("ord", "A"))

    assert vm.registers[1] == 65


def test_stdlib_chr():
    vm = execute_program_helper(helper_stdlib("chr", 65))

    addr = vm.registers[1]
    assert vm.load_memory(addr) == 1
    assert vm.load_memory(addr + 1) == 65


def test_stdlib_size():
    vm = execute_program_helper(helper_stdlib("size", "hello"))

    assert vm.registers[1] == 5


def test_stdlib_substring():
    vm = execute_program_helper(helper_stdlib("substring", "hello", 1, 2))

    addr = vm.registers[1]
    assert vm.load_memory(addr) == 2
    assert vm.load_memory(addr + 1) == ord("e")
    assert vm.load_memory(addr + 2) == ord("l")


def test_stdlib_concat():
    vm = execute_program_helper(helper_stdlib("concat", "ab", "cd"))

    addr = vm.registers[1]
    assert vm.load_memory(addr) == 4
    assert vm.load_memory(addr + 1) == ord("a")
    assert vm.load_memory(addr + 2) == ord("b")
    assert vm.load_memory(addr + 3) == ord("c")
    assert vm.load_memory(addr + 4) == ord("d")


def test_stdlib_tstrcmp():
    vm = execute_program_helper(helper_stdlib("tstrcmp", "a", "b"))

    assert from_u16(vm.registers[1]) < 0


def test_stdlib_div():
    vm = execute_program_helper(helper_stdlib("div", 42, 6))

    assert vm.registers[1] == 7


def test_stdlib_mod():
    vm = execute_program_helper(helper_stdlib("mod", 10, 3))

    assert vm.registers[1] == 1


def test_stdlib_not():
    vm = execute_program_helper(helper_stdlib("not", 1))

    assert vm.registers[1] == 0


def test_stdlib_getchar_ord():
    with patch("sys.stdin", StringIO("abc\n")):
        vm = main(["test/assets/cs350/getchar_ord.hera"])

    assert vm.registers[1] == ord("a")


def test_stdlib_getchar(capsys):
    with patch("sys.stdin", StringIO("abc\n")):
        vm = main(["test/assets/cs350/getchar.hera"])

    addr = vm.registers[1]
    assert vm.memory[addr] == 1
    assert vm.memory[addr + 1] == ord("a")


def test_stdlib_putchar_ord(capsys):
    execute_program_helper(helper_stdlib("putchar_ord", 98))

    assert capsys.readouterr().out == "b"


def test_stdlib_flush():
    # Just making sure it doesn't raise any SystemExit exceptions.
    execute_program_helper(helper_stdlib("flush"))


def test_stdlib_printint(capsys):
    execute_program_helper(helper_stdlib("printint", -42))

    assert capsys.readouterr().out == "-42"


def test_stdlib_printbool(capsys):
    execute_program_helper(helper_stdlib("printbool", 1))

    assert capsys.readouterr().out == "true"


def test_stdlib_print(capsys):
    execute_program_helper(helper_stdlib("print", "hello"))

    assert capsys.readouterr().out == "hello"


def test_stdlib_println(capsys):
    execute_program_helper(helper_stdlib("println", "hello"))

    assert capsys.readouterr().out == "hello\n"


def test_stdlib_ungetchar():
    vm = execute_program_helper(helper_stdlib("ungetchar"))

    assert vm.input_pos == -1


def test_stdlib_getline():
    with patch("sys.stdin", StringIO("hello\n")):
        vm = main(["test/assets/cs350/getline.hera"])

    addr = vm.registers[1]
    assert vm.memory[addr] == 5
    assert vm.memory[addr + 1] == ord("h")
    assert vm.memory[addr + 2] == ord("e")
    assert vm.memory[addr + 3] == ord("l")
    assert vm.memory[addr + 4] == ord("l")
    assert vm.memory[addr + 5] == ord("o")


def test_stdlib_reg_ord():
    vm = execute_program_helper(helper_stdlib_register("ord", "A"))

    assert vm.registers[1] == 65


def test_stdlib_reg_chr():
    vm = execute_program_helper(helper_stdlib_register("chr", 65))

    addr = vm.registers[1]
    assert vm.load_memory(addr) == 1
    assert vm.load_memory(addr + 1) == 65


def test_stdlib_reg_size():
    vm = execute_program_helper(helper_stdlib_register("size", "hello"))

    assert vm.registers[1] == 5


@pytest.mark.skip("not implemented in HERA-C")
def test_stdlib_reg_substring():
    vm = execute_program_helper(helper_stdlib_register("substring", "hello", 1, 2))

    addr = vm.registers[1]
    assert vm.load_memory(addr) == 2
    assert vm.load_memory(addr + 1) == ord("e")
    assert vm.load_memory(addr + 2) == ord("l")


@pytest.mark.skip("not implemented in HERA-C")
def test_stdlib_reg_concat():
    vm = execute_program_helper(helper_stdlib_register("concat", "ab", "cd"))

    addr = vm.registers[1]
    assert vm.load_memory(addr) == 4
    assert vm.load_memory(addr + 1) == ord("a")
    assert vm.load_memory(addr + 2) == ord("b")
    assert vm.load_memory(addr + 3) == ord("c")
    assert vm.load_memory(addr + 4) == ord("d")


def test_stdlib_reg_tstrcmp():
    vm = execute_program_helper(helper_stdlib_register("tstrcmp", "a", "b"))

    assert from_u16(vm.registers[1]) < 0


def test_stdlib_reg_div():
    vm = execute_program_helper(helper_stdlib_register("div", 42, 6))

    assert vm.registers[1] == 7


def test_stdlib_reg_mod():
    vm = execute_program_helper(helper_stdlib_register("mod", 10, 3))

    assert vm.registers[1] == 1


def test_stdlib_reg_not():
    vm = execute_program_helper(helper_stdlib_register("not", 1))

    assert vm.registers[1] == 0


def test_stdlib_reg_getchar_ord():
    with patch("sys.stdin", StringIO("abc\n")):
        vm = main(["test/assets/cs350/getchar_ord_reg.hera"])

    assert vm.registers[1] == ord("a")


def test_stdlib_reg_getchar(capsys):
    with patch("sys.stdin", StringIO("abc\n")):
        vm = main(["test/assets/cs350/getchar_reg.hera"])

    addr = vm.registers[1]
    assert vm.memory[addr] == 1
    assert vm.memory[addr + 1] == ord("a")


def test_stdlib_reg_putchar_ord(capsys):
    execute_program_helper(helper_stdlib_register("putchar_ord", 98))

    assert capsys.readouterr().out == "b"


def test_stdlib_reg_flush():
    # Just making sure it doesn't raise any SystemExit exceptions.
    execute_program_helper(helper_stdlib_register("flush"))


def test_stdlib_reg_printint(capsys):
    execute_program_helper(helper_stdlib_register("printint", -42))

    assert capsys.readouterr().out == "-42"


def test_stdlib_reg_printbool(capsys):
    execute_program_helper(helper_stdlib_register("printbool", 1))

    assert capsys.readouterr().out == "true"


def test_stdlib_reg_print(capsys):
    execute_program_helper(helper_stdlib_register("print", "hello"))

    assert capsys.readouterr().out == "hello"


def test_stdlib_reg_println(capsys):
    execute_program_helper(helper_stdlib_register("println", "hello"))

    assert capsys.readouterr().out == "hello\n"


def test_stdlib_reg_ungetchar():
    vm = execute_program_helper(helper_stdlib_register("ungetchar"))

    assert vm.input_pos == -1


def test_stdlib_reg_getline():
    with patch("sys.stdin", StringIO("hello\n")):
        vm = main(["test/assets/cs350/getline_reg.hera"])

    addr = vm.registers[1]
    assert vm.memory[addr] == 5
    assert vm.memory[addr + 1] == ord("h")
    assert vm.memory[addr + 2] == ord("e")
    assert vm.memory[addr + 3] == ord("l")
    assert vm.memory[addr + 4] == ord("l")
    assert vm.memory[addr + 5] == ord("o")


def helper_stdlib(f, *args):
    """Construct a HERA program that calls the Tiger function named `f` with the given
    arguments, using the parameters-on-the-stack calling convention.
    """
    data = ""
    for i, arg in enumerate(args):
        if isinstance(arg, str):
            data += "DLABEL(arg{})\nLP_STRING({})\n".format(i, json.dumps(arg))

    preamble = """\
#include <Tiger-stdlib-stack-data.hera>

CBON()
MOVE(R12, SP)
INC(SP, {})
"""
    preamble = preamble.format(len(args) + 3)

    load_args = ""
    for i, arg in enumerate(args):
        if isinstance(arg, str):
            load_args += "SET(R1, arg{})\nSTORE(R1, {}, R12)\n".format(i, i + 3)
        else:
            load_args += "SET(R1, {})\nSTORE(R1, {}, R12)\n".format(arg, i + 3)

    epilogue = """\
CALL(R12, {})
LOAD(R1, 3, R12)
DEC(SP, {})
HALT()

#include <Tiger-stdlib-stack.hera>
    """
    epilogue = epilogue.format(f, len(args) + 3)

    return data + preamble + load_args + epilogue


def helper_stdlib_register(f, *args):
    """Construct a HERA program that calls the Tiger function named `f` with the given
    arguments, using the parameters-in-registers calling convention.
    """
    data = ""
    for i, arg in enumerate(args):
        if isinstance(arg, str):
            data += "DLABEL(arg{})\nLP_STRING({})\n".format(i, json.dumps(arg))

    preamble = """\
#include <Tiger-stdlib-reg-data.hera>

CBON()
MOVE(R12, SP)
"""

    load_args = ""
    for i, arg in enumerate(args):
        if isinstance(arg, str):
            load_args += "SET(R{}, arg{})\n".format(i + 1, i)
        else:
            load_args += "SET(R{}, {})\n".format(i + 1, arg)

    epilogue = """\
CALL(R12, {})
HALT()

#include <Tiger-stdlib-reg.hera>
    """
    epilogue = epilogue.format(f)

    return data + preamble + load_args + epilogue



import pytest
import re
from io import StringIO
from unittest.mock import patch

from hera.data import Settings, VOLUME_VERBOSE
from hera.main import dump_state, main
from hera.vm import VirtualMachine


def test_main_non_existent_file(capsys):
    with pytest.raises(SystemExit):
        main(["unicorn.hera"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == 'Error: file "unicorn.hera" does not exist.\n'


def test_main_preprocess_non_existent_file(capsys):
    with pytest.raises(SystemExit):
        main(["preprocess", "unicorn.hera"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == 'Error: file "unicorn.hera" does not exist.\n'


def test_main_debug(capsys):
    # Just making sure we can invoke the debugger from the command-line, not testing
    # any of the debugger's functionality.
    with patch("sys.stdin", StringIO("quit")):
        main(["debug", "test/assets/cs240/factorial.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""


def test_main_with_short_version_flag(capsys):
    with pytest.raises(SystemExit):
        main(["-v"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert re.match(r"^hera-py [0-9.]+ for HERA version [0-9.]+$", captured.out)


def test_main_with_long_version_flag(capsys):
    with pytest.raises(SystemExit):
        main(["--version"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert re.match(r"^hera-py [0-9.]+ for HERA version [0-9.]+$", captured.out)


def test_main_with_version_flag_and_other_flags(capsys):
    with pytest.raises(SystemExit):
        main(["--version", "main.hera"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert (
        captured.err == "--version may not be combined with other flags or commands.\n"
    )


def test_main_with_help_flag_and_other_flags(capsys):
    with pytest.raises(SystemExit):
        main(["--help", "main.hera"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == "--help may not be combined with other flags or commands.\n"


def test_main_with_multiple_positional_args(capsys):
    with pytest.raises(SystemExit):
        main(["a.hera", "b.hera"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == "Too many file paths supplied.\n"


def test_main_with_no_positional_args(capsys):
    with pytest.raises(SystemExit):
        main([])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == "No file path supplied.\n"


def test_main_with_short_quiet_flag(capsys):
    with patch("sys.stdin", StringIO("SET(R1, 42)")):
        main(["-q", "-"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == "\n"


def test_main_with_long_quiet_flag(capsys):
    with patch("sys.stdin", StringIO("SET(R1, 42)")):
        main(["--quiet", "-"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == "\n"


def test_main_with_quiet_and_verbose_flags(capsys):
    with pytest.raises(SystemExit):
        main(["--quiet", "--verbose", "main.hera"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == "--quiet and --verbose are incompatible.\n"


def test_main_preprocess_and_debug_with_big_stack_flag(capsys):
    with pytest.raises(SystemExit):
        main(["--big-stack", "preprocess", "main.hera"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == "--big-stack is not compatible with the chosen mode.\n"


def test_main_preprocess_with_warn_return_off_flag(capsys):
    with pytest.raises(SystemExit):
        main(["--warn-return-off", "preprocess", "main.hera"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == "--warn-return-off is not compatible with the chosen mode.\n"


def test_main_with_verbose_flag(capsys):
    with patch("sys.stdin", StringIO("SET(R1, 42)")):
        main(["--verbose", "-"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert (
        captured.err
        == """\


Virtual machine state after execution (HERA v2.4.0):
    R1  = 0x002a = 42 = '*'
    R2  = 0x0000 = 0
    R3  = 0x0000 = 0
    R4  = 0x0000 = 0
    R5  = 0x0000 = 0
    R6  = 0x0000 = 0
    R7  = 0x0000 = 0
    R8  = 0x0000 = 0
    R9  = 0x0000 = 0
    R10 = 0x0000 = 0
    R11 = 0x0000 = 0
    R12 = 0x0000 = 0
    R13 = 0x0000 = 0
    R14 = 0x0000 = 0
    R15 = 0x0000 = 0

    Carry-block flag is OFF
    Carry flag is OFF
    Overflow flag is OFF
    Zero flag is OFF
    Sign flag is OFF
"""
    )


def test_main_with_big_stack_flag(capsys):
    program = "DLABEL(X)\nSET(R1, X)"
    with patch("sys.stdin", StringIO(program)):
        main(["--big-stack", "-"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert "R1  = 0xc167" in captured.err


def test_main_with_no_debug_flag(capsys):
    with pytest.raises(SystemExit):
        program = "print_reg(R1)"
        with patch("sys.stdin", StringIO(program)):
            main(["--no-debug-ops", "-"])

    captured = capsys.readouterr()
    assert (
        captured.err
        == """\

Error: debugging instructions disallowed with --no-debug-ops flag, line 1 col 1 of <stdin>

  print_reg(R1)
  ^

"""
    )


def test_main_with_code_flag(capsys):
    with pytest.raises(SystemExit):
        main(["--code", "whatever.hera"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == "--code is not compatible with the chosen mode.\n"


def test_main_with_data_flag(capsys):
    with pytest.raises(SystemExit):
        main(["--data", "whatever.hera"])

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == "--data is not compatible with the chosen mode.\n"


def test_no_ANSI_color_when_stderr_is_not_tty():
    buf = StringIO()
    with patch("sys.stderr", buf):
        with patch("sys.stdin", StringIO(")")):
            with pytest.raises(SystemExit):
                main(["-"])

    assert (
        buf.getvalue()
        == """\

Error: expected HERA operation or #include, line 1 col 1 of <stdin>

  )
  ^

"""
    )


def test_main_with_init_flag(capsys):
    with patch("sys.stdin", StringIO("NOP()")):
        main(["-", "--init=r1=4, r2=5"])

    captured = capsys.readouterr().err
    assert "R1  = 0x0004 = 4" in captured
    assert "R2  = 0x0005 = 5" in captured


def test_main_with_init_flag_withg_different_syntax(capsys):
    with patch("sys.stdin", StringIO("NOP()")):
        main(["-", "--init", "r1=4, r2=5"])

    captured = capsys.readouterr().err
    assert "R1  = 0x0004 = 4" in captured
    assert "R2  = 0x0005 = 5" in captured


def test_main_with_dash_to_separate_arguments(capsys):
    with pytest.raises(SystemExit):
        main(["debug", "--", "--version"])

    captured = capsys.readouterr()
    assert captured.out == ""
    # Make sure we get a "file not found" error and not a "cannot use --version with
    # debug subcommand" error, which would indicate that the dash separator was not
    # being interpreted correctly.
    assert 'file "--version" does not exist' in captured.err


def test_credits_flag(capsys):
    with pytest.raises(SystemExit):
        main(["--credits"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert "Ian Fisher" in captured.out


def test_throttle_flag(capsys):
    with patch("sys.stdin", StringIO("LABEL(start); BR(start)")):
        main(["--throttle", "100", "-"])

    captured = capsys.readouterr()
    assert "Program throttled after 100 instructions." in captured.err


def test_dump_state(capsys):
    dump_state(VirtualMachine(), Settings(volume=VOLUME_VERBOSE))

    assert (
        capsys.readouterr().err
        == """\

Virtual machine state after execution (HERA v2.4.0):
    R1  = 0x0000 = 0
    R2  = 0x0000 = 0
    R3  = 0x0000 = 0
    R4  = 0x0000 = 0
    R5  = 0x0000 = 0
    R6  = 0x0000 = 0
    R7  = 0x0000 = 0
    R8  = 0x0000 = 0
    R9  = 0x0000 = 0
    R10 = 0x0000 = 0
    R11 = 0x0000 = 0
    R12 = 0x0000 = 0
    R13 = 0x0000 = 0
    R14 = 0x0000 = 0
    R15 = 0x0000 = 0

    Carry-block flag is OFF
    Carry flag is OFF
    Overflow flag is OFF
    Zero flag is OFF
    Sign flag is OFF
"""
    )



from hera.data import Constant, DataLabel, Label, Token
from hera.op import (
    check_in_range,
    check_label,
    check_register,
    check_register_or_label,
    check_string,
    SET,
)


def test_op_to_string():
    assert str(SET(R(1), SYM("top"))) == "SET(R1, top)"


def test_op_to_string_with_integer():
    assert str(SET(R(1), INT(12))) == "SET(R1, 12)"


def test_check_register_with_non_register():
    assert check_register(INT(10)) == "expected register"


def test_check_register_with_program_counter():
    assert (
        check_register(SYM("PC"))
        == "program counter cannot be accessed or changed directly"
    )


def test_check_register_with_valid_registers():
    assert check_register(R(0)) is None
    assert check_register(R(1)) is None
    assert check_register(R(12)) is None


def test_check_register_or_label_with_integer():
    assert check_register_or_label(INT(10), {}) == "expected register or label"


def test_check_register_or_label_with_undefined_symbol():
    assert check_register_or_label(SYM("n"), {}) == "undefined symbol"


def test_check_register_or_label_with_constant():
    err = check_register_or_label(SYM("n"), {"n": Constant(1)})
    assert err == "constant cannot be used as label"


def test_check_register_or_label_with_data_label():
    err = check_register_or_label(SYM("n"), {"n": DataLabel(1)})
    assert err == "data label cannot be used as branch label"


def test_check_register_or_label_with_valid_args():
    assert check_register_or_label(R(7), {}) is None
    assert check_register_or_label(SYM("n"), {"n": Label(1)}) is None


def test_check_label_with_non_label():
    assert check_label(R(1)) == "expected label"


def test_check_label_with_valid_label():
    assert check_label(SYM("l")) is None


def test_check_string_with_non_string():
    assert check_string(R(7)) == "expected string literal"
    assert check_string(SYM("hello")) == "expected string literal"


def test_check_string_with_valid_string():
    assert check_string(STR("hello")) is None


def test_check_in_range_with_undefined_symbol():
    assert check_in_range(SYM("n"), {}, lo=0, hi=128) == "undefined constant"


def test_check_in_range_with_label():
    err = check_in_range(SYM("n"), {"n": Label(1)}, lo=0, hi=128)
    assert err == "cannot use label as constant"


def test_check_in_range_with_data_label():
    err = check_in_range(SYM("n"), {"n": DataLabel(1)}, lo=0, hi=128)
    assert err == "cannot use label as constant"


def test_check_in_range_with_non_integer():
    assert check_in_range(R(1), {}, lo=0, hi=128) == "expected integer"


def test_check_in_range_with_out_of_range_integer():
    err = check_in_range(INT(-1), {}, lo=0, hi=128)
    assert err == "integer must be in range [0, 128)"


def test_check_in_range_with_another_out_of_range_integer():
    err = check_in_range(INT(128), {}, lo=0, hi=128)
    assert err == "integer must be in range [0, 128)"


def test_check_in_range_with_out_of_range_symbol():
    err = check_in_range(SYM("n"), {"n": Constant(128)}, lo=0, hi=128)
    assert err == "integer must be in range [0, 128)"


def test_check_in_range_with_valid_integers():
    assert check_in_range(INT(50), {}, lo=0, hi=128) is None
    assert check_in_range(INT(0), {}, lo=0, hi=128) is None
    assert check_in_range(INT(127), {}, lo=0, hi=128) is None
    assert check_in_range(SYM("n"), {"n": Constant(127)}, lo=0, hi=128) is None


def R(i):
    return Token(Token.REGISTER, i)


def SYM(s):
    return Token(Token.SYMBOL, s)


def STR(s):
    return Token(Token.STRING, s)


def INT(x):
    return Token(Token.INT, x)



import pytest
from .utils import helper

from hera.data import Settings
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine(Settings(color=False))


def test_exec_print_reg(vm, capsys):
    vm.registers[1] = 70

    helper(vm, "print_reg(R1)")

    assert capsys.readouterr().out == "R1 = 0x0046 = 70 = 'F'\n"


def test_print_reg_increments_pc(vm):
    helper(vm, "print_reg(R1)")

    assert vm.pc == 1


def test_exec_print(vm, capsys):
    helper(vm, 'print("Hello, world!")')

    assert capsys.readouterr().out == "Hello, world!"


def test_print_increments_pc(vm):
    helper(vm, 'print("Hello, world!")')

    assert vm.pc == 1


def test_exec_println(vm, capsys):
    helper(vm, 'println("Hello, world!")')

    assert capsys.readouterr().out == "Hello, world!\n"


def test_println_increments_pc(vm):
    helper(vm, 'println("Hello, world!")')

    assert vm.pc == 1


def test___eval_increments_pc(vm):
    helper(vm, '__eval("0")')

    assert vm.pc == 1



"""
The parser for the debugger's expression mini-language.

Author:  Ian Fisher (iafisher@fastmail.com)
Version: January 2019
"""
from collections import namedtuple

from ..lexer import Lexer
from ..data import HERAError, Token
from ..utils import register_to_index


def parse(line: str) -> "SeqNode":
    """
    Return a parse tree for the line of code. Raise a SyntaxError if it is not well-
    formatted.
    """
    return MiniParser(Lexer(line)).parse()


class MiniParser:
    """
    A parser for the debugger's expression mini-language. The abstract grammar below
    describes the language. Arithmetic operations have the usual precedence.

      start := FORMAT? exprlist

      exprlist := (expr COMMA)* expr

      expr := expr op expr | LPAREN expr RPAREN | MINUS expr | AT expr | atom
      op   := PLUS | MINUS | ASTERISK | SLASH
      atom := REGISTER | INT | SYMBOL

    Based on the Pratt parser from Thorsten Ball's "Writing an Interpreter in Go".
    https://interpreterbook.com/

    All match methods in this class expect that self.lexer.tkn is equal to the first
    token of the expression to be matched, and they leave self.lexer.tkn on the first
    token of the next expression.
    """

    def __init__(self, lexer: Lexer) -> None:
        self.lexer = lexer

    def parse(self) -> "SeqNode":
        tree = self.match_exprlist()
        if self.lexer.tkn.type == Token.EOF:
            return tree
        else:
            raise SyntaxError("trailing input")

    def match_exprlist(self) -> "SeqNode":
        """Match a sequence of comma-separated expressions."""
        seq = []

        if self.lexer.tkn.type == Token.FMT:
            fmt = self.lexer.tkn.value
            self.lexer.next_token()
        else:
            fmt = ""

        while True:
            expr = self.match_expr(PREC_LOWEST)
            seq.append(expr)
            if self.lexer.tkn.type == Token.COMMA:
                self.lexer.next_token()
            else:
                break

        return SeqNode(fmt, seq)

    def match_expr(self, precedence: int) -> "AbstractNode":
        """Parse the expression with the given precedence."""
        tkn = self.lexer.tkn
        # This line is solely to satisfy mypy.
        left = AbstractNode()
        if tkn.type == Token.AT:
            self.lexer.next_token()
            address = self.match_expr(PREC_PREFIX)
            left = MemoryNode(address)
        elif tkn.type == Token.INT:
            try:
                left = IntNode(int(tkn.value, base=0))
            except ValueError:
                raise SyntaxError("invalid integer literal: {}".format(tkn))
            else:
                self.lexer.next_token()
        elif tkn.type == Token.MINUS:
            self.lexer.next_token()
            left = PrefixNode("-", self.match_expr(PREC_PREFIX))
        elif tkn.type == Token.REGISTER:
            try:
                left = RegisterNode(register_to_index(tkn.value))
            except HERAError:
                raise SyntaxError("{} is not a valid register".format(tkn.value))
            self.lexer.next_token()
        elif tkn.type == Token.SYMBOL:
            left = SymbolNode(tkn.value)
            self.lexer.next_token()
        elif tkn.type == Token.LPAREN:
            self.lexer.next_token()
            left = self.match_expr(PREC_LOWEST)
            if self.lexer.tkn.type != Token.RPAREN:
                self.unexpected(self.lexer.tkn)
            self.lexer.next_token()
        else:
            self.unexpected(tkn)

        infix_tkn = self.lexer.tkn
        while infix_tkn.type in PREC_MAP and precedence < PREC_MAP[infix_tkn.type]:
            infix_precedence = PREC_MAP[infix_tkn.type]
            self.lexer.next_token()
            right = self.match_expr(infix_precedence)
            left = InfixNode(infix_tkn.value, left, right)
            infix_tkn = self.lexer.tkn
        return left

    def unexpected(self, tkn: Token) -> None:
        if tkn.type == Token.EOF:
            raise SyntaxError("premature end of input")
        elif tkn.type == Token.UNKNOWN:
            raise SyntaxError("unrecognized input `{}`".format(tkn))
        else:
            raise SyntaxError("did not expect `{}` in this position".format(tkn))


class AbstractNode:
    """An abstract base class for AST nodes. Used for type annotations."""


class SeqNode(namedtuple("SeqNode", ["fmt", "seq"]), AbstractNode):
    def __str__(self):
        seqstr = ", ".join(map(str, self.seq))
        if self.fmt:
            return ":{} {}".format(self.fmt, seqstr)
        else:
            return seqstr


class MemoryNode(namedtuple("MemoryNode", ["address"]), AbstractNode):
    def __str__(self):
        return "@{}".format(wrap(self.address))


class RegisterNode(namedtuple("RegisterNode", ["value"]), AbstractNode):
    def __str__(self):
        return "R" + str(self.value)


class IntNode(namedtuple("IntNode", ["value"]), AbstractNode):
    def __str__(self):
        return str(self.value)


class SymbolNode(namedtuple("SymbolNode", ["value"]), AbstractNode):
    def __str__(self):
        return self.value


class PrefixNode(namedtuple("PrefixNode", ["op", "arg"]), AbstractNode):
    def __str__(self):
        return "{}{}".format(self.op, wrap(self.arg))


class InfixNode(namedtuple("InfixNode", ["op", "left", "right"]), AbstractNode):
    def __str__(self):
        return "{} {} {}".format(wrap(self.left), self.op, wrap(self.right))


def wrap(node: AbstractNode) -> str:
    """Stringify the node and wrap it in parentheses if necessary."""
    if isinstance(node, InfixNode):
        return "(" + str(node) + ")"
    else:
        return str(node)


# Operator precedence
PREC_MAP = {Token.PLUS: 1, Token.MINUS: 1, Token.SLASH: 2, Token.ASTERISK: 2}
PREC_LOWEST = 0
PREC_PREFIX = 3



from io import StringIO
from unittest.mock import patch

from hera.main import main


def test_disassemble_from_stdin(capsys):
    with patch("sys.stdin", StringIO("e1ff")):
        main(["disassemble", "-"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert captured.out == "\nSETLO(R1, 255)\n"


def test_disassemble_set_inc(capsys):
    main(["disassemble", "test/assets/asm/set_inc.hera.lcode"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        captured.out
        == """\
SETLO(R1, 255)
SETHI(R3, 42)
INC(R10, 1)
DEC(R1, 20)
"""
    )


def test_disassemble_binary_op(capsys):
    main(["disassemble", "test/assets/asm/binary_op.hera.lcode"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        captured.out
        == """\
AND(R1, R7, R12)
OR(R0, R4, R2)
ADD(R3, R5, R8)
SUB(R9, R13, R15)
MUL(R1, R2, R3)
XOR(R8, R7, R4)
"""
    )


def test_disassemble_branch(capsys):
    main(["disassemble", "test/assets/asm/branch.hera.lcode"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        captured.out
        == """\
BR(R1)
BL(R7)
BGE(R1)
BLE(R1)
BG(R1)
BULE(R1)
BUG(R1)
BZ(R1)
BNZ(R1)
BC(R1)
BNC(R1)
BS(R1)
BNS(R1)
BV(R1)
BNV(R1)
CALL(R12, R11)
RETURN(R12, R13)
"""
    )


def test_disassemble_relative_branch(capsys):
    main(["disassemble", "test/assets/asm/rel_branch.hera.lcode"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        captured.out
        == """\
BRR(1)
BLR(255)
BGER(100)
BLER(156)
BGR(1)
BULER(1)
BUGR(1)
BZR(1)
BNZR(1)
BCR(1)
BNCR(1)
BSR(1)
BNSR(1)
BVR(1)
BNVR(1)
"""
    )


def test_disassemble_flag(capsys):
    main(["disassemble", "test/assets/asm/flag.hera.lcode"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        captured.out
        == """\
SAVEF(R5)
RSTRF(R5)
FON(4)
FOFF(5)
FSET5(3)
FSET4(2)
"""
    )


def test_disassemble_shift(capsys):
    main(["disassemble", "test/assets/asm/shift.hera.lcode"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        captured.out
        == """\
LSL(R3, R4)
LSR(R5, R6)
LSL8(R7, R8)
LSR8(R8, R9)
ASL(R1, R2)
ASR(R2, R3)
"""
    )


def test_disassemble_misc(capsys):
    main(["disassemble", "test/assets/asm/misc.hera.lcode"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        captured.out
        == """\
LOAD(R1, 10, R14)
STORE(R2, 0, R14)
SWI(10)
RTI()
"""
    )


def test_disassemble_another_branch(capsys):
    with patch("sys.stdin", StringIO("0011")):
        main(["disassemble", "-"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert captured.out == "\nBRR(17)\n"
